'''
Date: 2024-12-18 18:35:52
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-12-18 18:43:31
Description: 
'''
import os

from pymol import cmd

if __name__ == '__main__':
    from lazydock.utils import uuid4
else:
    from ..utils import uuid4

def get_seq(pose: str, fasta: bool = False):
    if os.path.isfile(pose):
        if fasta:
            pose_name = os.path.basename(pose).split('.')[0]
        else:
            pose_name = uuid4()
        cmd.load(pose, pose_name)
        pose = pose_name
    seq = cmd.get_fastastr(pose)
    if fasta:
        return seq
    return ''.join(seq.split('\n')[1:])


if __name__ == '__main__':
    print(get_seq('data_tmp/pdb/RECEPTOR.pdb', False))


# Copyright (C) 2014 E. Prabhu Raman prabhu@outerbanks.umaryland.edu
#
# Modified 11/6/2018 by Justin Lemkul to add lone pair support
# needed for CGenFF >= 4.0 halogens
#
# Modified 01/10/2019 by Conrard Tetsassi to work with Networkx 2.3
# Included notes on bonds, angles, and dihedrals
#
# For help/questions/bug reports, please contact Justin Lemkul jalemkul@vt.edu
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
# GNU Affero General Public License for more details.
# <http://www.gnu.org/licenses/>

# EXAMPLE: You have a drug-like molecule in drug.mol2 file
# ParamChem returns a CHARMM stream file drug.str with topology and parameters
# INPUT
# The program needs four inputs:
#	(1) The first argument (resname) is found in the RESI entry in the CHARMM stream file; for example
#		RESI DRUG		  0.000  ! here DRUG is the resname
#	(2) drug.mol2 is the .mol2 which you supplied to ParamChem
#	(3) drug.str contains the topology entry and CGenFF parameters that ParamChem generates for you
#	(4) charmm36.ff should contain the CHARMM force-field in GROMACS format
#		Download it from: http://mackerell.umaryland.edu/CHARMM_ff_params.html

# OUTPUT
# The program will generate 4 output files ("DRUG" is converted to lowercase and the files are named accordingly):
#	(1) drug.itp - contains GROMACS itp
#	(2) drug.prm - contains parameters obtained from drug.str which are converted to GROMACS format and units
#	(3) drug.top - A Gromacs topology file which incorporates (1) and (2)
#	(4) drug_ini.pdb - Coordinates of the molecule obtained from drug.mol2

# Modified 2024/12/16 by BHM-Bob_G, to fit networkx 3.X and run as a sub-package

import math
import os
import re

import networkx as nx
import numpy as np
from mbapy_lite.file import opts_file


#=================================================================================================================
def check_versions(str_filename,ffdoc_filename):
    ffver = 0	# CGenFF version in force field directory
    strver = 0	# CGenFF version in stream file
    for line in opts_file(str_filename, way='lines'):
        if line.startswith("* For use with CGenFF version"):
            entry = re.split('\s+', line.lstrip())
            strver = entry[6]
            print("--Version of CGenFF detected in ",str_filename,":",strver)
    for line in opts_file(ffdoc_filename, way='lines'):
        if line.startswith("Parameters taken from CHARMM36 and CGenFF"):
            entry = re.split('\s+', line.lstrip())
            ffver = entry[6]
            print("--Version of CGenFF detected in ",ffdoc_filename,":",ffver)

    # warn the user about version mismatch
    if strver != ffver:
        print("\nWARNING: CGenFF versions are not equivalent!\n")

    # in case something has gone horribly wrong
    if (strver == 0) or (ffver == 0):
        print("\nERROR: Could not detect CGenFF version. Exiting.\n")
        exit()

#-----------------------------------------------------------------------
## jal
def is_lp(s):
    if ((s[0]=='L') and (s[1]=='P')):
        return True
    return False

#-----------------------------------------------------------------------
## jal
def is_lp_host_atom(self,name):
    for ai in range (0,self.nvsites):
        if (name==self.G._node[ai]['at1']):
            return True
    return False

#-----------------------------------------------------------------------
## jal - only for COLINEAR lone pairs, since CGenFF only needs this now
def construct_lp(x1,y1,z1,x2,y2,z2,dist):
    dx = x1-x2
    dy = y1-y2
    dz = z1-z2
    dr = math.sqrt(dx*dx+dy*dy+dz*dz)
    dr = dist/dr
    # set LP coords
    xlp = x1+dr*dx
    ylp = y1+dr*dy
    zlp = z1+dr*dz

    return xlp,ylp,zlp

#-----------------------------------------------------------------------
## jal
def find_vsite(self, atnum):
    for i in range (0, self.nvsites):
        # if we find the LP host, find the LP atom index
        if (self.G._node[i]['at1'] == self.G._node[atnum]['name']):
            for j in range (0, self.natoms):
                if (self.G._node[i]['vsite'] == self.G._node[j]['name']):
                    return j

#-----------------------------------------------------------------------
def read_gmx_atomtypes(filename):
    atomtypes = []
    for line in opts_file(filename, way='lines'):
        if line.startswith(";"):
            continue
        if line == '\n':
            continue
        entry = re.split('\s+', line.lstrip())
        var = [entry[0],entry[1]]
        atomtypes.append(var)
    return atomtypes
#-----------------------------------------------------------------------
def get_filelist_from_gmx_forcefielditp(ffdir,ffparentfile):
    filelist=[]
    for line in opts_file(os.path.join(ffdir, ffparentfile), way='lines'):
        if line.startswith("#include"):
            entry = re.split('\s+', line.lstrip())
            filename = ffdir + "/" + entry[1].replace("\"","")
            filelist.append(filename)
    return filelist
#-----------------------------------------------------------------------
def read_gmx_anglpars(filename):

    angllines = []
    section="NONE"
    for line in opts_file(filename, way='lines'):
        if line.startswith(";"):
            continue
        if line.startswith("\n"):
            continue
        if line.startswith("["):
            section="NONE"
        if(section=="ANGL"):
            angllines.append(line)
        if line.startswith("[ angletypes ]"):
            section="ANGL"

    anglpars = []
    anglpar = {}
    for line in angllines:
        entry = re.split('\s+', line.lstrip())
        ai, aj, ak, eq = entry[0],entry[1],entry[2],float(entry[4])
        anglpars.append([ai,aj,ak,eq])

    return anglpars
#-----------------------------------------------------------------------
def get_charmm_rtp_lines(filename,molname):
    foundmol=0
    store=0
    rtplines=[]
    section="NONE"
    for line in opts_file(filename, way='lines'):
        if(store==1) and line.startswith("RESI"):
            store=0

        if line.startswith("RESI"):
            entry = re.split('\s+', line.lstrip())
            rtfmolname=entry[1]
            if(rtfmolname == molname):
                store=1

        if line.startswith("END"):
            store=0

        if(store==1):
            rtplines.append(line)

    return rtplines
#-----------------------------------------------------------------------
def get_charmm_prm_lines(filename):
    foundmol=0
    store=0
    prmlines=[]
    section="NONE"
    for line in opts_file(filename, way='lines'):

        if line.startswith("END"):
            section="NONE"
            store=0

        if(store):
            prmlines.append(line)

        if line.startswith("read para"):
            section="PRM"
            store=1


    return prmlines
#-----------------------------------------------------------------------
def parse_charmm_topology(rtplines):
    topology = {}
    noblanks = [x for x in rtplines if len(x.strip())>0]
    nocomments = [x for x in noblanks if x.strip()[0] not in ['*','!']]
    section = "BONDS"	# default
    state = "free"
    for line in nocomments:
        if state == "free":
            if line.find("MASS") == 0:
                if "ATOMS" not in list(topology.keys()):
                    topology["ATOMS"] = {}
                s = line.split()
                idx,name,mass,type = int(s[1]),s[2],float(s[3]),s[4]
                if line.find("!"):
                    comment = line[line.find("!")+1:].strip()
                else:
                    comment = ""
                topology["ATOMS"][name] = [idx,mass,type,comment]
            elif line.find("DECL") == 0:
                if "DECL" not in list(topology.keys()):
                    topology["DECL"] = []
                decl = line.split()[1]
                topology["DECL"].append(decl)
            elif line.find("DEFA") == 0:
                topology["DEFA"] = line[4:]
            elif line.find("AUTO") == 0:
                topology["AUTO"] = line[4:]
            elif line.find("RESI") == 0:
                if "RESI" not in list(topology.keys()):
                    topology["RESI"] = {}
                state = "resi"
                s = line.split()
                resname, charge = s[1],float(s[2])
                topology["RESI"][resname] = {}
                topology["RESI"][resname]["charge"] = charge
                topology["RESI"][resname]["cmaps"] = []
                topology["RESI"][resname]["vsites"] = []
                topology["RESI"][resname]["bonds"] = []
                topology["RESI"][resname]["impropers"] = []
                topology["RESI"][resname]["double_bonds"] = []
                group = -1
            elif line.find("PRES") == 0:
                state = "pres"
                s = line.split()
                presname, charge = s[1],float(s[2])
            elif line.find("END") == 0:
                return topology
        elif state == "resi":
            if line.find("RESI")==0:
                state = "resi"
                s = line.split()
                resname, charge = s[1],float(s[2])
                topology["RESI"][resname] = {}
                topology["RESI"][resname]["charge"] = charge
                topology["RESI"][resname]["cmaps"] = []
                topology["RESI"][resname]["vsites"] = []
                topology["RESI"][resname]["bonds"] = []
                topology["RESI"][resname]["impropers"] = []
                topology["RESI"][resname]["double_bonds"] = []
                #topology["RESI"][resname]["groups"] = []
                group = -1

            elif line.find("GROU")==0:
                group += 1
                topology["RESI"][resname][group] = []
            elif line.find("ATOM")==0:
                if line.find('!'):
                    line = line[:line.find('!')]
                s = line.split()
                name,type,charge = s[1],s[2],float(s[3])
                topology["RESI"][resname][group].append((name,type,charge))
            ## jal - adding lone pair support
            elif line.find("LONE")==0:
                if line.find('!'):
                    line = line[:line.find('!')]
                s = line.split()
                name,at1,at2,dist = s[2],s[3],s[4],(float(s[6])*0.1)
                topology["RESI"][resname]["vsites"].append((name,at1,at2,dist))
            elif line.find("BOND")==0:
                if line.find('!'):
                    line = line[:line.find('!')]
                s = line.split()
                nbond = (len(s)-1)/2
                for i in range(nbond):
                    p,q = s[1+2*i],s[2+2*i]
                    ## jal - ignore "bonds" to lone pairs
                    if ((is_lp(p)==False) and (is_lp(q)==False)):
                        topology["RESI"][resname]["bonds"].append((p,q))
            elif line.find("DOUB")==0:
                if line.find('!'):
                    line = line[:line.find('!')]
                s = line.split()
                ndouble = (len(s)-1)/2
                for i in range(ndouble):
                    p,q = s[1+2*i],s[2+2*i]
                    topology["RESI"][resname]["double_bonds"].append((p,q))
            elif line.find("IMPR")==0:
                if line.find('!'):
                    line = line[:line.find('!')]
                s = line.split()
                nimproper = (len(s)-1)/4
                for i in range(nimproper):
                    impr = s[1+4*i],s[2+4*i],s[3+4*i],s[4+4*i]
                    topology["RESI"][resname]["impropers"].append(impr)
            elif line.find("CMAP")==0:
                if line.find('!'):
                    line = line[:line.find('!')]
                s = line.split()
                #nimproper = (len(s)-1)/4
                #for i in range(nimproper):
                cmap = s[1:9]
                topology["RESI"][resname]["cmaps"].append(cmap)
            elif line.find("DONOR")==0:
                continue	# ignore for now
            elif line.find("ACCEPTOR")==0:
                continue
            elif line.find("IC")==0:
                continue

    return topology
#-----------------------------------------------------------------------
def parse_charmm_parameters(prmlines):

    parameters = {}
    cmapkey = ()
    noblanks = [x for x in prmlines if len(x.strip())>0]
    nocomments = [x for x in noblanks if x.strip()[0] not in ['*','!']]
    section = "ATOM"	# default
    for line in nocomments:
                #print line
        sectionkeys = [ "BOND", "ANGL", "DIHE", \
                "IMPR", "CMAP", "NONB", "HBON", "NBFI" ]
        key = line.split()[0]

                #exit()

        if key[0:4] in sectionkeys:
            section = key[0:4]
            continue

        if section not in list(parameters.keys()):
            parameters[section] = []

                #print line
        if section == "BOND":
            if line.find('!'):
                line = line[:line.find('!')]
            s = line.split()
            ai, aj, kij, rij = s[0],s[1],float(s[2]),float(s[3])
            parameters["BOND"].append((ai,aj,kij,rij))
        elif section == "ANGL":
            if line.find('!'):
                line = line[:line.find('!')]
            s = line.split()
            ai, aj, ak = s[0],s[1],s[2]
            other = list(map(float,s[3:]))
            parameters["ANGL"].append([ai,aj,ak]+other)
        elif section == "DIHE":
            if line.find('!'):
                line = line[:line.find('!')]
            s = line.split()
            ai, aj, ak, al, k, n, d = s[0],s[1],s[2],s[3],float(s[4]),int(s[5]),float(s[6])
            parameters["DIHE"].append([ai,aj,ak,al,k,n,d])
        elif section == "IMPR":
            if line.find('!'):
                line = line[:line.find('!')]
            s = line.split()
            ai, aj, ak, al, k, d = s[0],s[1],s[2],s[3],float(s[4]),float(s[6])
            parameters["IMPR"].append([ai,aj,ak,al,k,d])
        elif section == "CMAP":
            if line.find('!'):
                line = line[:line.find('!')]
            if cmapkey == ():
                s = line.split()
                a,b,c,d,e,f,g,h = s[0:8]
                N = int(s[8])
                cmapkey = (a,b,c,d,e,f,g,h,N)
                cmaplist = []
            else:
                cmapdata = list(map(float,line.split()))
                cmaplist += cmapdata
                if len(cmaplist) == N**2:
                    parameters["CMAP"].append([cmapkey,cmaplist])
                    cmapkey = ()
        elif section == "NONB":
            if line.find("cutnb")>=0 or line.find("wmin")>=0 or line.find("CUTNB")>=0 or line.find("WMIN")>=0 :
                continue
            bang = line.find('!')
            if bang>0:
                comment = line[bang+1:]
                prm = line[:bang].split()
            else:
                comment = ""
                prm = line.split()
            atname = prm[0]
            epsilon = -float(prm[2])
            half_rmin = float(prm[3])
            parameters["NONB"].append((atname,epsilon,half_rmin))

            if len(prm)>4:
                epsilon14 = -float(prm[5])
                half_rmin14 = float(prm[6])
                if "NONBONDED14" not in list(parameters.keys()):
                    parameters["NONBONDED14"] = []
                parameters["NONBONDED14"].append((atname,epsilon14,half_rmin14))


    return parameters
#-----------------------------------------------------------------------
def write_gmx_bon(parameters,header_comments,filename):
    kcal2kJ = 4.18400

    outp = open(filename,"w")
    outp.write("%s\n"%(header_comments))
    outp.write("[ bondtypes ]\n")
    kbond_conversion = 2.0*kcal2kJ/(0.1)**2		# [kcal/mol]/A**2 -> [kJ/mol]/nm**2
                        # factor of 0.5 because charmm bonds are Eb(r)=Kb*(r-r0)**2
    rbond_conversion = .1			# A -> nm
    outp.write(";%7s %8s %5s %12s %12s\n"%("i","j","func","b0","kb"))
    if("BOND" in parameters):
        for p in parameters["BOND"]:
            ai,aj,kij,rij = p
            rij *= rbond_conversion
            kij *= kbond_conversion
            outp.write("%8s %8s %5i %12.8f %12.2f\n"%(ai,aj,1,rij,kij))

    kangle_conversion = 2.0*kcal2kJ		# [kcal/mol]/rad**2 -> [kJ/mol]/rad**2
                                        # factor of 0.5 because charmm angles are Ea(r)=Ka*(a-a0)**2
    kub_conversion = 2.0*kcal2kJ/(0.1)**2		# [kcal/mol]/A**2 -> [kJ/mol]/nm**2prm
    ub0_conversion = 0.1						# A -> nm

    outp.write("\n\n[ angletypes ]\n")
    outp.write(";%7s %8s %8s %5s %12s %12s %12s %12s\n"\
            %("i","j","k","func","theta0","ktheta","ub0","kub"))
    if("ANGL" in parameters):
        for p in parameters["ANGL"]:
            if len(p) == 5:
                ai,aj,ak,kijk,theta = p
                kub = 0.0
                ub0 = 0.0
            else:
                ai,aj,ak,kijk,theta,kub,ub0 = p

            kijk *= kangle_conversion
            kub *= kub_conversion
            ub0 *= ub0_conversion
            outp.write("%8s %8s %8s %5i %12.6f %12.6f %12.8f %12.2f\n"\
                    %(ai,aj,ak,5,theta,kijk,ub0,kub))

    kdihe_conversion = kcal2kJ
    outp.write("\n\n[ dihedraltypes ]\n")
    outp.write(";%7s %8s %8s %8s %5s %12s %12s %5s\n"\
            %("i","j","k","l","func","phi0","kphi","mult"))
    #parameters["DIHEDRALS"].sort(demote_wildcards)
    if("DIHE" in parameters):
        for p in parameters["DIHE"]:
            ai,aj,ak,al,k,n,d = p
            k *= kdihe_conversion
            outp.write("%8s %8s %8s %8s %5i %12.6f %12.6f %5i\n"\
                    %(ai,aj,ak,al,9,d,k,n))

    kimpr_conversion = kcal2kJ*2	# see above
    outp.write("\n\n[ dihedraltypes ]\n")
    outp.write("; 'improper' dihedrals \n")
    outp.write(";%7s %8s %8s %8s %5s %12s %12s\n"\
            %("i","j","k","l","func","phi0","kphi"))
    if("IMPR" in parameters):
    #parameters["IMPROPERS"].sort(demote_wildcards)
        for p in parameters["IMPR"]:
            ai,aj,ak,al,k,d = p
            k *= kimpr_conversion
            outp.write("%8s %8s %8s %8s %5i %12.6f %12.6f\n"\
                    %(ai,aj,ak,al,2,d,k))

    outp.close()
    return
#-----------------------------------------------------------------------
def write_gmx_mol_top(filename,ffdir,prmfile,itpfile,molname):
    outp = open(filename,"w")
    outp.write("#include \"%s/forcefield.itp\"\n" % (ffdir))
    outp.write("\n")
    outp.write("; additional params for the molecule\n")
    outp.write("#include \"%s\"\n" % (prmfile))
    outp.write("\n")
    outp.write("#include \"%s\"\n" % (itpfile))
    outp.write("\n")
    outp.write("#include \"%s/tip3p.itp\"\n" % (ffdir))
    outp.write("#ifdef POSRES_WATER\n")
    outp.write("; Position restraint for each water oxygen\n")
    outp.write("[ position_restraints ]\n")
    outp.write(";  i funct		 fcx		fcy		   fcz\n")
    outp.write("   1	1		1000	   1000		  1000\n")
    outp.write("#endif\n")
    outp.write("\n")
    outp.write("; Include topology for ions\n")
    outp.write("#include \"%s/ions.itp\"\n" % (ffdir))
    outp.write("\n")
    outp.write("[ system ]\n")
    outp.write("; Name\n")
    outp.write("mol\n")
    outp.write("\n")
    outp.write("[ molecules ]\n")
    outp.write("; Compound		  #mols\n")
    outp.write("%s			1\n" % (molname))
    outp.write("\n")

    outp.close()
#=================================================================================================================
class atomgroup:
    """
    A class that contains the data structures and functions to store and process
    data related to groups of atoms (read molecules)

    USAGE: m = atomgroup()
    """

    def __init__(self):
        self.G = nx.Graph()
        self.name = ""
        self.natoms = 0
        self.nvsites = 0
        self.nbonds = 0
        self.angles = []
        self.nangles = 0
        self.dihedrals = []
        self.ndihedrals = 0
        self.impropers = []
        self.nimpropers = 0
        #self.coord=np.zeros((self.natoms,3),dtype=float)

    #-----------------------------------------------------------------------
    def read_charmm_rtp(self,rtplines,atomtypes):
        """
        Reads CHARMM rtp
        Reads atoms, bonds, impropers
        Stores connectivity as a graph
        Autogenerates angles and dihedrals

        USAGE: m = atomgroup() ; m.read_charmm_rtp(rtplines,atomtypes)

        """
        #initialize everything
        self.G = nx.Graph()
        self.name = ""
        self.natoms = 0
        self.nvsites = 0
        self.nbonds = 0
        self.angles = []
        self.nangles = 0
        self.dihedrals = []
        self.ndihedrals = 0
        self.impropers = []
        self.nimpropers = 0

        atm = {}

        for line in rtplines:
            if line.find('!'):
                line = line[:line.find('!')]

                if line.startswith("RESI"):
                    entry = re.split('\s+', line.lstrip())
                    self.name=entry[1]

                if line.startswith("ATOM"):
                    entry = re.split('\s+', line.lstrip())
                    atm[self.natoms] = {'type':entry[2], 'resname':self.name, 'name':entry[1],
                          'charge':float(entry[3]),'mass':float(0.00), 'beta':float(0.0),
                            'x':float(9999.9999),'y':float(9999.9999),'z':float(9999.9999),'segid':self.name, 'resid':'1' }

                    for typei in atomtypes:
                        if(typei[0] == atm[self.natoms]['type']):
                            atm[self.natoms]['mass'] = float(typei[1])
                            break

                    #self.G.add_node(self.natoms, atm[self.natoms])
                    #@@@@@@@@@@@@@@ Conrard modification @@@@@@@@@@@
                    att = self.natoms # attribute
                    nodd = {att:atm[self.natoms]} # node to add
                    self.G.add_node(att) #instanciate the attribute
                    nx.set_node_attributes(self.G, nodd)
                    self.natoms=self.natoms+1

                ## jal - adding lone pair support
                if line.startswith("LONE"):
                    #entry = re.split('\s+', line.rstrip(line.lstrip()))
                    entry = re.split('\s+', line.lstrip())
                    atm[self.nvsites] = {'vsite':entry[2], 'at1':entry[3], 'at2':entry[4],
                            'dist':(float(entry[6])*0.1),
                            'x':float(9999.9999),'y':float(9999.9999),'z':float(9999.9999) }
                    # DEBUG
                    # print "Found lone pair in RTF: %s %s %s %.3f\n" % (atm[self.nvsites]['vsite'], atm[self.nvsites]['at1'], atm[self.nvsites]['at2'], atm[self.nvsites]['dist'])

                    # jal - as above
                    att = self.nvsites
                    nodd = {att:atm[self.nvsites]}
                    self.G.add_node(att)
                    nx.set_node_attributes(self.G, nodd)
                    self.nvsites=self.nvsites+1

                if line.startswith("BOND") or line.startswith("DOUB"):
                    #entry = re.split('\s+', line.rstrip(line.lstrip()))
                    entry = re.split('\s+', line.lstrip())
                    numbonds = int((len(entry)-1)/2)
                    for bondi in range(0,numbonds):
                        found1 = False
                        found2 = False
                        for i in range(0,self.natoms):
                            if(atm[i]['name'] == entry[(bondi*2)+1]):
                                found1 = True
                                break
                        for j in range(0,self.natoms):
                            if(atm[j]['name'] == entry[(bondi*2)+2]):
                                found2 = True
                                break
                        if(not found1):
                            print("Error:atomgroup:read_charmm_rtp> Atomname not found in top",entry[(bondi*2)+1])
                        if(not found2):
                            print("Error:atomgroup:read_charmm_rtp> Atomname not found in top",entry[(bondi*2)+2])
                        ## jal - ignore "bonds" to lone pairs
                        if ((is_lp(atm[i]['name'])==False) and (is_lp(atm[j]['name'])==False)):
                            self.G.add_edge(i,j)
                            self.G[i][j]['order']='1' # treat all bonds as single for now
                            self.nbonds=self.nbonds+1

                if line.startswith("IMP"):
                    entry = re.split('\s+', line.lstrip())
                    numimpr = int((len(entry)-2)/4)
                    for impi in range(0,numimpr):
                        for i in range(0,self.natoms):
                            if(atm[i]['name'] == entry[(impi*4)+1]):
                                break
                        for j in range(0,self.natoms):
                            if(atm[j]['name'] == entry[(impi*4)+2]):
                                break
                        for k in range(0,self.natoms):
                            if(atm[k]['name'] == entry[(impi*4)+3]):
                                break
                        for l in range(0,self.natoms):
                            if(atm[l]['name'] == entry[(impi*4)+4]):
                                break
                        var = [i,j,k,l]
                        self.impropers.append(var)

        self.nimpropers = len(self.impropers)
        if(self.ndihedrals > 0 or self.nangles > 0):
            print("WARNING:atomgroup:read_charmm_rtp> Autogenerating angl-dihe even though they are preexisting",self.nangles,self.ndihedrals)
        self.autogen_angl_dihe()
        self.coord = np.zeros((self.natoms,3),dtype=float)
#-----------------------------------------------------------------------
    def autogen_angl_dihe(self):
        self.angles = []
        for atomi in range(0,self.natoms):
            nblist = []
            for nb in self.G.neighbors(atomi):
                nblist.append(nb)
            for i in range(0,len(nblist)-1):
                for j in range(i+1,len(nblist)):
                    var = [nblist[i],atomi,nblist[j]]
                    self.angles.append(var)
        self.nangles = len(self.angles)
        self.dihedrals = []
        for i,j in self.G.edges():
            nblist1 = []
            for nb in self.G.neighbors(i):
                if(nb != j):
                    nblist1.append(nb)
            nblist2 = []
            for nb in self.G.neighbors(j):
                if(nb != i):
                    nblist2.append(nb)
            if(len(nblist1) > 0 and len(nblist2) > 0 ):
                for ii in range(0,len(nblist1)):
                    for jj in range(0,len(nblist2)):
                        var = [nblist1[ii],i,j,nblist2[jj]]
                        if(var[0] != var[3]):
                            self.dihedrals.append(var)
        self.ndihedrals = len(self.dihedrals)
#-----------------------------------------------------------------------
    def get_nonplanar_dihedrals(self,angl_params):
        nonplanar_dihedrals=[]
        cutoff=179.9
        for var in self.dihedrals:
            d1=self.G._node[var[0]]['type']
            d2=self.G._node[var[1]]['type']
            d3=self.G._node[var[2]]['type']
            d4=self.G._node[var[3]]['type']
            keep=1
            for angl_param in angl_params:
                p1=angl_param[0]
                p2=angl_param[1]
                p3=angl_param[2]
                eq=angl_param[3]
                if( d2==p2 and ( ( d1==p1 and d3==p3) or (d1==p3 and d3==p1))):
                    if(eq > cutoff):
                        keep=-1
                        break
                if( d3==p2 and ( (d2==p1 and d4==p3) or (d2==p3 and d4==p1))):
                    if(eq > cutoff):
                        keep=-1
                        break

            if(keep==1):
                nonplanar_dihedrals.append(var)

        return nonplanar_dihedrals
#-----------------------------------------------------------------------
    def write_gmx_itp(self,filename,angl_params):
        f = open(filename, 'w')
        f.write("; Created by cgenff_charmm2gmx.py\n")
        f.write("\n")
        f.write("[ moleculetype ]\n")
        f.write("; Name			   nrexcl\n")
        f.write("%s				 3\n" % self.name)
        f.write("\n")
        f.write("[ atoms ]\n")
        f.write(";	 nr		  type	resnr residue  atom   cgnr	   charge		mass  typeB    chargeB		massB\n")
        f.write("; residue	 1 %s rtp %s q	qsum\n" % (self.name,self.name))
        pairs14 = nx.Graph()
        for atomi in range(0,self.natoms):
            pairs14.add_node(atomi)
            f.write("%6d %10s %6s %6s %6s %6d %10.3f %10.3f   ;\n" %
               ( atomi+1,self.G._node[atomi]['type'],
               self.G._node[atomi]['resid'],self.name,self.G._node[atomi]['name'],atomi+1,
               self.G._node[atomi]['charge'],self.G._node[atomi]['mass'] ) )
        f.write("\n")
        f.write("[ bonds ]\n")
        f.write(";	ai	  aj funct			  c0			c1			  c2			c3\n")
        for i,j in self.G.edges():
            #f.write("%5d %5d	 1\n" % (i+1,j+1) )
            f.write("%5d %5d	 1 ;  %10s %10s\n" % (i+1,j+1,self.G._node[i]['type'],self.G._node[j]['type']) )
        f.write("\n")
        f.write("[ pairs ]\n")
        f.write(";	ai	  aj funct			  c0			c1			  c2			c3\n")
        for var in self.dihedrals:
            if (len(nx.dijkstra_path(self.G,var[0],var[3])) == 4): #this is to remove 1-2 and 1-3 included in dihedrals of rings
                pairs14.add_edge(var[0],var[3])
        for i,j in pairs14.edges():
            f.write("%5d %5d	 1\n" % (i+1,j+1) )
            #f.write("%5d %5d	 1 ;  %10s %10s\n" % (i+1,j+1,self.G._node[i]['type'],self.G._node[j]['type']) )
            ## jal - add LP pairs, same as parent atom
            ## Use is_lp_host_atom() to test each index, then find associated vsite
            if ((is_lp_host_atom(self,self.G._node[i]['name'])==True)):
                k = find_vsite(self, i)
                f.write("%5d %5d	 1\n" % (k+1,j+1) )
            if ((is_lp_host_atom(self,self.G._node[j]['name'])==True)):
                k = find_vsite(self, j)
                f.write("%5d %5d	 1\n" % (k+1,i+1) )
        f.write("\n")
        f.write("[ angles ]\n")
        f.write(";	ai	  aj	ak funct			c0			  c1			c2			  c3\n")
        for var in self.angles:
            f.write("%5d %5d %5d	5 ; %10s %10s %10s\n" % (var[0]+1,var[1]+1,var[2]+1,\
                self.G._node[var[0]]['type'],self.G._node[var[1]]['type'],self.G._node[var[2]]['type']) )
        f.write("\n")
        f.write("[ dihedrals ]\n")
        f.write(";	ai	  aj	ak	  al funct			  c0			c1			  c2			c3			  c4			c5\n")
        nonplanar_dihedrals=self.get_nonplanar_dihedrals(angl_params)
        for var in nonplanar_dihedrals:
            f.write("%5d %5d %5d %5d	 9 ; %10s %10s %10s %10s\n" % (var[0]+1,var[1]+1,var[2]+1,var[3]+1,\
            self.G._node[var[0]]['type'],self.G._node[var[1]]['type'],self.G._node[var[2]]['type'],self.G._node[var[3]]['type']) )
        f.write("\n")
        if(self.nimpropers > 0):
            f.write("[ dihedrals ]\n")
            f.write(";	ai	  aj	ak	  al funct			  c0			c1			  c2			c3\n")
            for var in self.impropers:
                f.write("%5d %5d %5d %5d	 2\n" % (var[0]+1,var[1]+1,var[2]+1,var[3]+1) )
            f.write("\n")
        ## jal - add vsite directive
        ## we use 2fd construction, introduced in GROMACS-2020
        if (self.nvsites > 0):
            func=2
            f.write("[ virtual_sites2 ]\n")
            f.write("; Site   from				funct a\n")
            for atomi in range (0,self.nvsites):
                vsite = 0
                at1 = 0
                at2 = 0
                # find atom name matches
                for ai in range (0, self.natoms):
                    if (self.G._node[ai]['name'] == self.G._node[atomi]['vsite']):
                        vsite = ai
                    if (self.G._node[ai]['name'] == self.G._node[atomi]['at1']):
                        at1 = ai
                    if (self.G._node[ai]['name'] == self.G._node[atomi]['at2']):
                        at2 = ai
                dist=self.G._node[atomi]['dist']*-1  # invert sign for GROMACS convention
                f.write("%5d %5d %5d %5d %8.3f\n" % (vsite+1, at1+1, at2+1, func, dist))
            f.write("\n")

        ## jal - add exclusions for vsite
        if (self.nvsites > 0):
            f.write("[ exclusions ]\n")
            f.write(";	ai	  aj\n")
            ## jal - explicitly add all 1-2, 1-3, and 1-4 exclusions
            ## for the lone pair, which are the same as 1-2, 1-3, 1-4
            ## exclusions for the host (bonds, angles, pairs)
            # first, exclude any LP from its host
            for i in range (0, self.natoms):
                if ((is_lp_host_atom(self,self.G._node[i]['name'])==True)):
                    # find the LP attached to this host, not necessarily consecutive
                    # in the topology
                    j = find_vsite(self, i)
                    f.write("%5d %5d	 1\n" % (i+1,j+1) )
            # first neighbors: 1-2
            for i,j in self.G.edges():
                if ((is_lp_host_atom(self,self.G._node[i]['name'])==True)):
                    k = find_vsite(self, i)
                    f.write("%5d %5d	 1\n" % (k+1,j+1) )
                if ((is_lp_host_atom(self,self.G._node[j]['name'])==True)):
                    k = find_vsite(self, j)
                    f.write("%5d %5d	 1\n" % (k+1,i+1) )
            # second neighbors: 1-3
            for var in self.angles:
                # only need to consider ends of the angle, not middle atom
                ai = var[0]
                ak = var[2]
                if ((is_lp_host_atom(self,self.G._node[ai]['name'])==True)):
                    l = find_vsite(self, ai)
                    f.write("%5d %5d	 1\n" % (l+1,ak+1) )
                if ((is_lp_host_atom(self,self.G._node[ak]['name'])==True)):
                    l = find_vsite(self, ak)
                    f.write("%5d %5d	 1\n" % (l+1,ai+1) )
            # third neighbors: 1-4
            for i,j in pairs14.edges():
                if ((is_lp_host_atom(self,self.G._node[i]['name'])==True)):
                    k = find_vsite(self, i)
                    f.write("%5d %5d	 1\n" % (k+1,j+1) )
                if ((is_lp_host_atom(self,self.G._node[j]['name'])==True)):
                    k = find_vsite(self, j)
                    f.write("%5d %5d	 1\n" % (k+1,i+1) )
            f.write("\n")

        f.close()

#-----------------------------------------------------------------------
    def read_mol2_coor_only(self,filename):
        check_natoms = 0
        check_nbonds = 0
        atm = {}
        section="NONE"
        for line in opts_file(filename, way='lines'):
            secflag=False
            if line.startswith("@"):
                secflag=True
                section="NONE"

            if((section=="NATO") and (not secflag)):
                entry = re.split('\s+', line.lstrip())
                check_natoms=int(entry[0])
                check_nbonds=int(entry[1])
                if(check_natoms != self.natoms):
                    # jal - if there are lone pairs, these will not be in the mol2 file
                    if (self.nvsites == 0):
                        print("Error in atomgroup.py: read_mol2_coor_only: no. of atoms in mol2 (%d) and top (%d) are unequal" % (check_natoms, self.natoms))
                        print("Usually this means the specified residue name does not match between str and mol2 files")
                        #print check_natoms,self.natoms
                        exit()
                    else:
                        print("")
                        print("NOTE 5: %d lone pairs found in topology that are not in the mol2 file. This is not a problem, just FYI!\n" % (self.nvsites))
                # jal - if we have correctly ignored bonds to LP then there is no need
                # for any check here
                if(check_nbonds != self.nbonds):
                    print("Error in atomgroup.py: read_mol2_coor_only: no. of bonds in mol2 (%d) and top (%d) are unequal" % (check_nbonds, self.nbonds))
                    #print check_nbonds,self.nbonds
                    exit()

                section="NONE"

            if((section=="MOLE") and (not secflag)):
                self.name=line.strip()
                section="NATO" #next line after @<TRIPOS>MOLECULE contains atom, bond numbers

            if((section=="ATOM") and (not secflag)):
                entry = re.split('\s+', line.lstrip())
                ## guard against blank lines
                if (len(entry) > 1):
                    ## jal - if there are lone pairs, these are not in mol2
                    ## and are not necessarily something we can just tack on at the
                    ## end of the coordinate section. Here, check the atom to see if it is
                    ## the first constructing atom, and if so, we put in a dummy LP entry.
                    atomi = int(entry[0])-1
                    self.G._node[atomi]['x'] = float(entry[2])
                    self.G._node[atomi]['y'] = float(entry[3])
                    self.G._node[atomi]['z'] = float(entry[4])
                    self.coord[atomi][0] = float(entry[2])
                    self.coord[atomi][1] = float(entry[3])
                    self.coord[atomi][2] = float(entry[4])
                    ## jal - if we have an atom that is the host for a LP, insert
                    ## the LP into the list
                    if (is_lp_host_atom(self,self.G._node[atomi]['name'])):
                        atomj = find_vsite(self, atomi)
                        # insert dummy entry for LP
                        self.G._node[atomj]['x'] = float(9999.99)
                        self.G._node[atomj]['y'] = float(9999.99)
                        self.G._node[atomj]['z'] = float(9999.99)
                        self.coord[atomj][0] = float(9999.99)
                        self.coord[atomj][1] = float(9999.99)
                        self.coord[atomj][2] = float(9999.99)

            if line.startswith("@<TRIPOS>MOLECULE"):
                section="MOLE"
            if line.startswith("@<TRIPOS>ATOM"):
                section="ATOM"
            if line.startswith("@<TRIPOS>BOND"):
                section="BOND"
#-----------------------------------------------------------------------
    def write_pdb(self,f):
        for atomi in range(0,self.natoms):
            if(len(self.G._node[atomi]['name']) > 4):
                print("error in atomgroup.write_pdb(): atom name > 4 characters")
                exit()
            ## jal - construct LP sites
            if (is_lp(self.G._node[atomi]['name'])):
                # DEBUG
                # print "Found LP in write_pdb: %s\n" % self.G._node[atomi]['name']
                # find constructing atoms, get their coordinates and construction distance*10
                atn1 = "dum"
                atn2 = "dum"
                dist = 0
                # loop over vsites
                for ai in range (0,self.nvsites):
                    if (self.G._node[ai]['vsite'] == self.G._node[atomi]['name']):
                        atn1 = self.G._node[ai]['at1']	 # atom name
                        atn2 = self.G._node[ai]['at2']	 # atom name
                        dist = self.G._node[ai]['dist']*10 # Angstrom for PDB, was saved as *0.1 for GMX

                # get atom indices
                at1 = 0
                at2 = 0
                for ai in range (0, self.natoms):
                    if (self.G._node[ai]['name'] == atn1):
                        at1 = ai
                    if (self.G._node[ai]['name'] == atn2):
                        at2 = ai

                # in case of failure
                if ((at1==0) and (at2==0)):
                    print("Failed to match LP-constructing atoms in write_pdb!\n")
                    exit()

                # DEBUG
                # print "Found LP in write_pdb: %d %s %s with dist: %.3f\n" % ((atomi+1,at1+1,at2+1,dist))

                # at1, at2, and dist only exist in vsite structure!
                x1=self.coord[at1][0]
                y1=self.coord[at1][1]
                z1=self.coord[at1][2]
                x2=self.coord[at2][0]
                y2=self.coord[at2][1]
                z2=self.coord[at2][2]

                xlp,ylp,zlp = construct_lp(x1,y1,z1,x2,y2,z2,dist)
                self.coord[atomi][0] = xlp
                self.coord[atomi][1] = ylp
                self.coord[atomi][2] = zlp
            f.write("%-6s%5d %-4s %-4s%5s%12.3f%8.3f%8.3f%6.2f%6.2f\n" %
                ("ATOM",atomi+1,self.G._node[atomi]['name'],self.name,self.G._node[atomi]['resid'],self.coord[atomi][0],
                self.coord[atomi][1],self.coord[atomi][2],1.0,self.G._node[atomi]['beta']))
        f.write("END\n")

#=================================================================================================================

def run_transform(mol_name: str, mol2_path: str, str_path: str, ffdir: str,
                  output_dir: str = None):
    """
    Parameters:
        - mol_name (str): name of the molecule
        - mol2_path (str): name of the mol2 file
        - str_path (str): name of the rtp file
        - ffdir (str): directory containing the forcefield files
        - output_dir (str): directory to write the output files to (default: None, save into parent directory of mol2_path)
    """
    atomtypes_path = os.path.join(ffdir, "atomtypes.atp")

    # for output
    output_dir = output_dir or os.path.dirname(os.path.abspath(mol2_path))
    itpfile = os.path.join(output_dir, mol_name.lower() + ".itp")
    prmfile = os.path.join(output_dir, mol_name.lower() + ".prm")
    initpdbfile = os.path.join(output_dir, mol_name.lower() + "_ini.pdb")
    topfile = os.path.join(output_dir, mol_name.lower() +".top")

    atomtypes = read_gmx_atomtypes(atomtypes_path)

    angl_params = []  # needed for detecting triple bonds
    filelist = get_filelist_from_gmx_forcefielditp(ffdir, "forcefield.itp")
    for filename in filelist:
        anglpars = read_gmx_anglpars(filename)
        angl_params = angl_params + anglpars


    m = atomgroup()
    rtplines = get_charmm_rtp_lines(str_path, mol_name)
    m.read_charmm_rtp(rtplines, atomtypes)


    m.read_mol2_coor_only(mol2_path)
    f = open(initpdbfile, 'w')
    m.write_pdb(f)
    f.close()


    prmlines = get_charmm_prm_lines(str_path)
    params = parse_charmm_parameters(prmlines)
    write_gmx_bon(params, "", prmfile)
    anglpars = read_gmx_anglpars(prmfile)
    angl_params = angl_params + anglpars # append the new angl params


    m.write_gmx_itp(itpfile, angl_params)
    write_gmx_mol_top(topfile, ffdir, prmfile, itpfile, mol_name)
    

if __name__ == '__main__':
    run_transform("LIG", r"data_tmp\gmx\LIG.mol2", r"data_tmp\gmx\LIG.str", r"data_tmp\gmx\charmm36-jul2022.ff")


'''
Date: 2024-08-16 09:36:38
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-12-15 19:34:35
Description: LazyDock Pymol Plugin
'''

import os

os.environ['MBAPY_FAST_LOAD'] = 'True'

import sys

sys.path.append(os.path.dirname(__file__))

from pymol import cmd

# do all init in frist level code block
import lazydock.pml_plugin._init_ as lazydock_pml_plugin_init
# GUILauncher is the main function of the plugin, so just init in this file.
from lazydock.pml_plugin.main import GUILauncher


def __init__(self):
    try:
        from pymol.plugins import addmenuitemqt
        addmenuitemqt('LazyDock', GUILauncher)
        return
    except Exception as e:
        print(e)
    self.menuBar.addmenuitem('Plugin', 'command', 'lazydock',
                             label = 'LazyDock', command = lambda s=self : GUILauncher(s)) 
    



'''
Date: 2024-11-27 17:01:54
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-12-02 17:00:54
Description: 
'''
import argparse
import os
import shutil
from pathlib import Path
from typing import Dict, List

from mbapy_lite.file import get_paths_with_extension, opts_file
from pymol import cmd

if __name__ == '__main__':
    from lazydock.scripts._script_utils_ import clean_path, show_args
    from lazydock.web.gen_pocket import (get_pocket_box_from_ProteinPlus,
                                         parse_pocket_box_from_ProteinPlus)
else:
    from ..web.gen_pocket import (get_pocket_box_from_ProteinPlus,
                                  parse_pocket_box_from_ProteinPlus)
    from ._script_utils_ import clean_path, show_args
    

def main(sys_args: List[str] = None):
    args_paser = argparse.ArgumentParser(description = 'Get pocket residues from ProteinPlus web server.')
    args_paser.add_argument('-r', '--receptor', type = str, default='.',
                            help='input mol2 file or dir path. Default is %(default)s.')
    args_paser.add_argument('-l', '--ligand', type = str, default=None,
                            help='ligand sdf file path. Default is %(default)s.')
    args_paser.add_argument('-o', '--output', type=str, default=None,
                            help='output dir. Default is %(default)s.')
    args_paser.add_argument('-m', '--method', type=str, default='extend', choices=['extend', 'mean'],
                            help='pocket calculation method. Default is %(default)s.')
    args = args_paser.parse_args(sys_args)
    # process IO path
    args.receptor = clean_path(args.receptor)
    if os.path.isfile(args.receptor):
        receptor_paths = [args.receptor]
    elif os.path.isdir(args.receptor):
        receptor_paths = get_paths_with_extension(args.receptor, ['.pdb'], recursive=args.recursive)
    else:
        raise ValueError(f'Input path {args.receptor} is not a pdb file or directory.')
    if not receptor_paths:
        raise ValueError(f'No pdb file found in {args.receptor}.')
    # show args
    show_args(args, ['receptor', 'ligand', 'output'])
    # sort bonds
    for receptor_path in receptor_paths:
        print(f'Getting pocket box from {receptor_path} ...')
        output_dir = Path(receptor_path).parent / f'{Path(receptor_path).stem}_pocket_box'
        if output_dir.exists():
            print(f'Output dir {output_dir} already exists, skip.')
            continue
        output_dir.mkdir(exist_ok=True, parents=True)
        shutil.copy(receptor_path, output_dir) # so that the zip file is in the same dir with pdb file
        # get pocket box
        get_pocket_box_from_ProteinPlus(os.path.join(output_dir, Path(receptor_path).name), ligand_path=args.ligand)
        # parse pocket box
        zip_path = get_paths_with_extension(output_dir, ['.zip'], recursive=False)[0]
        for index in [[0], [1], [2], [0, 1], [0, 1, 2]]:
            idx_str = ','.join(map(str, index))
            if args.method == 'extend':
                pocket = parse_pocket_box_from_ProteinPlus(zip_path, index, True, method=args.method)
            elif args.method == 'mean':
                df, pocket = parse_pocket_box_from_ProteinPlus(zip_path, index, True, method=args.method)
                df.to_excel(output_dir / f"pocket_{idx_str}_box.xlsx", index=False)
            # save pocket residues
            opts_file(output_dir / f"pocket_{idx_str}_residues.json", 'w', way='json', data=pocket)
            cmd.save(output_dir / f"pocket_{idx_str}_box.pse")
            print(f'Pocket residues saved to {output_dir}.')


if __name__ == "__main__":
    main()


'''
Date: 2024-11-20 21:45:38
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-11-23 19:46:20
Description: transfer from http://www.mdtutorials.com/gmx/complex/Files/sort_mol2_bonds.pl
'''

import os
import sys

from mbapy_lite.file import opts_file


def sort_bonds(mol2: str) -> str:
    if os.path.exists(mol2):
        mol2_lines = opts_file(mol2).split('\n')
    else:
        mol2_lines = mol2.split('\n')
    # Check for header lines
    if not mol2_lines[0].startswith('@<TRIPOS>'):
        sys.exit(f"Nonstandard header found: {mol2_lines[0]}. Please delete header lines until the TRIPOS molecule definition.")
    # output
    result_lines = []
    # Get number of atoms and bonds from mol2 file
    natom, nbond = map(int, mol2_lines[2].strip().split()[:2])
    print(f"Found {natom} atoms in the molecule, with {nbond} bonds.")
    # Print out everything up until the bond section
    i = 0
    while not mol2_lines[i].startswith('@<TRIPOS>BOND'):
        result_lines.append(mol2_lines[i])
        i += 1
    # Print the bond section header line to output
    result_lines.append(mol2_lines[i])
    i += 1
    # Read in the bonds and sort them
    bondfmt = "%6d%6d%6d%5s"
    tmparray = []
    for j in range(nbond):
        parts = mol2_lines[i + j].split()
        ai, aj = int(parts[1]), int(parts[2])
        # Reorder if second atom number < first
        if aj < ai:
            ai, aj = aj, ai
        tmparray.append((ai, aj, int(parts[0]), parts[3]))
    # Sort the bonds
    tmparray.sort()
    # Loop over tmparray to find each atom number
    nbond = 0
    for x in range(1, natom + 1):
        bondarray = [t for t in tmparray if t[0] == x]
        if bondarray:
            for y, (ai, aj, bi, bj) in enumerate(sorted(bondarray, key=lambda x: x[1])):
                nbond += 1
                result_lines.append(bondfmt % (bi, ai, aj, bj))
    return '\n'.join(result_lines)


if __name__ == "__main__":
    # dev code
    sorted_mol2 = sort_bonds('data_tmp/pdb/ligand.mol2')
    opts_file('data_tmp/pdb/ligand_fix.mol2', 'w', data = sorted_mol2)
    print(sorted_mol2)


'''
Date: 2024-12-13 20:18:59
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-12-25 21:19:19
Description: steps most from http://www.mdtutorials.com/gmx
'''

import argparse
import os
import shutil
from typing import Dict, List, Tuple, Union

from mbapy_lite.base import put_err
from mbapy_lite.file import get_dir, get_paths_with_extension
from tqdm import tqdm

from lazydock.scripts._script_utils_ import clean_path, show_args


def main(sys_args: List[str] = None):
    args_paser = argparse.ArgumentParser(description = 'Get pocket residues from ProteinPlus web server.')
    args_paser.add_argument('-sr', '--source-root', type = str,
                            help='root to the files or directory to be distributed.')
    args_paser.add_argument('-sn', '--source-name', type = str,
                            help='sub string in source files or directory name.')
    args_paser.add_argument('-st', '--source-type', type = str,
                            help='file type in source directory, multiple types can be separated by comma.')
    args_paser.add_argument('-m', '--mode', type = str, choice=['file', 'dir'], default='file',
                            help='source search mode, file or directory. Default is %(default)s.')
    args_paser.add_argument('-dr', '--dist-root', type = str,
                            help='path to the dist root directory.')
    args_paser.add_argument('-dn', '--dist-name', type = str,
                            help='sub string in dist directory name.')
    args_paser.add_argument('-in', '--item-name', type = str, default='',
                            help='sub string in dist directory item files name. Default is %(default)s.')
    args_paser.add_argument('-it', '--item-type', type = str, default='',
                            help='file type in dist directory, multiple types can be separated by comma. Default is %(default)s.')
    args_paser.add_argument('-n', '--new-name', type = str, default=None,
                            help='sub string in file name. Default is %(default)s.')
    args = args_paser.parse_args(sys_args)
    # process IO path
    args.source_root = clean_path(args.source_root)
    args.source_type = args.source_type.split(',')
    args.dist_root = clean_path(args.dist_root)
    args.dist_type = args.dist_type.split(',')
    # search source files or directories
    if args.mode == 'file':
        sources = get_paths_with_extension(args.source_root, args.source_type, name_substr=args.source_name)
    else:
        sources = get_dir(args.source_root, 0, None, args.source_type, True, args.source_name)
    if not sources:
        return put_err(f'No dist directory found in {args.source}, exit.')
    # search dist directories
    dist_dirs = get_dir(args.dist_root, file_extensions=args.dist_type, recursive=True, dir_name_substr=args.dist_name)
    if not dist_dirs:
        return put_err(f'No dist directory found in {args.dist_root}, exit.')
    # show args
    show_args(args, ['source_root', 'source_name', 'source_type', 'mode', 'dist_root',
                     'dist_name', 'dist_type', 'item_name', 'item_type', 'new_name'])
    # copy file or directory to dist directory
    for dist_dir in tqdm(dist_dirs, total=len(dist_dirs)):
        for source in sources:
            dist_name = os.path.basename(source) if args.new_name is None else args.new_name
            dist_path = os.path.join(dist_dir, dist_name)
            if args.mode == 'file':
                shutil.copy(source, dist_path)
            else:
                shutil.copytree(source, dist_path)


if __name__ == "__main__":
    main()


'''
Date: 2024-11-02 22:09:17
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-11-03 16:17:45
Description: 
'''
import io
import sys
from typing import List

from nicegui import ui
from pymol import CmdException, api, cmd


class LazyPml:
    def __init__(self, app, _dev_mode: bool = False):
        
        self._app = app
        
        sys.stdout = self.captured_output = io.StringIO()
        
        self.cmd_string = ''
        self.ui_log = None
        
    def build_cmd_gui(self):
        with ui.row().classes('w-full'):
            ui.input(label='pymol python api call').bind_value_to(self, 'cmd_string').classes('flex flex-grow')
            ui.button('execute', on_click=self.execute_cmd).classes('flex flex-grow')
        with ui.row():
            self.ui_log = ui.log()
            
    def execute_cmd(self):
        try:
            exec(self.cmd_string, globals(), locals())
            self.ui_log.push(self.captured_output.getvalue().split('\n')[-2])
        except CmdException as e:
            self.ui_log.push(str(e))
        
    def build_gui(self):
        with ui.tabs().classes('w-full').props('align=left active-bg-color=blue') as tabs:
            self.ui_cmd_tab = ui.tab('command').props('no-caps')
        with ui.tab_panels(tabs, value=self.ui_cmd_tab).classes('w-full'):
            with ui.tab_panel(self.ui_cmd_tab):
                self.build_cmd_gui()
        # return self
        return self
            
        
# dev mode
if __name__ in {"__main__", "__mp_main__"}:    
    cmd.reinitialize()
    cmd.load('data_tmp/pdb/LIGAND.pdb', 'ligand')
    cmd.load('data_tmp/pdb/RECEPTOR.pdb', 'receptor')
    
    from main import GUILauncher
    GUILauncher()


import os
import platform
import re
import tempfile
import time
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from mbapy_lite.base import put_err
from mbapy_lite.file import opts_file
from mbapy_lite.web import TaskPool
from pymol import cmd
from tqdm import tqdm

if __name__ == '__main__':
    from lazydock import config
    from lazydock.pml.interaction_utils import sort_func
    from lazydock.utils import uuid4
else:
    from .. import config
    from ..utils import uuid4
    from .interaction_utils import sort_func
    
    
def parse_ligplus_output(result_path: str) -> Dict[str, List[Tuple[Tuple[int, str, str], Tuple[int, str, str], float]]]:
    """
    parse single LigPlus output files (ligplus.hhb or ligplus.nnb) to get atom-level interactions between receptor and ligand.
    
    file format:
    RRR.C.NNNNI.AAAA.....RRR.C.NNNNI.AAAA....DDDD
    
    Parameters:
        - result_path (str): LigPlus output file path, either ligplus.hhb or ligplus.nnb
    Returns:
        interactions (list): ((receptor_resi, receptor_resn, receptor_chain), (ligand_resi, ligand_resn, ligand_chain), distance)
    """
    lines = re.findall(r'([A-Z]+) ([A-Z]) + ([0-9]+) +(\w+) +([A-Z]+) ([A-Z]) + ([0-9]+) +(\w+) +([0-9\.]+)\n', opts_file(result_path))
    return [((int(line[2]), line[0], line[1]), (int(line[6]), line[4], line[5]), float(line[-1])) for line in lines]

    
def _run_ligplus_for_complex(ligplus_dir: str, complex_pdbstr: str,
                             receptor_chain: str, ligand_chain: str, mode: List[str], cutoff: float,
                             force_cwd: bool = False, w_dir: str = None, **kwargs) -> Dict[str, List[Tuple[Tuple[int, str, str], Tuple[int, str, str], float]]]:
    """
    run LigPlus for a complex, assume comonents.cif is in the LigPlus params directory
    command lines refer to https://github.com/eachanjohnson/dimpyplot/blob/master/dimpyplot.py
    """
    # set ligplus lib path
    pwd_cmd = 'pwd'
    if platform.platform().startswith('Windows'):
        ligplus_lib = os.path.join(ligplus_dir, 'lib', 'exe_win')
        pwd_cmd = 'cd'
    elif platform.platform().startswith('Linux'):
        ligplus_lib = os.path.join(ligplus_dir, 'lib', 'exe_linux')
    else:
        put_err(f'Unsupported platform: {platform.platform()}', _exit=True)
    ligplus_params = os.path.join(ligplus_dir, 'lib', 'params')
    # run LigPlus in temporary directory
    w_dir_handle = None
    if w_dir is None:
        w_dir_handle = tempfile.TemporaryDirectory()
        w_dir = w_dir_handle.name + '/' # because hbplus just append file name to w_dir
    else:
        w_dir = os.path.abspath(w_dir) + '/'
        os.makedirs(w_dir, exist_ok=True)
    cwd_cmd = f'cd "{w_dir}" && {pwd_cmd} &&'
    if force_cwd:
        cwd_cmd = ''
        now_wd = os.getcwd()
        os.chdir(w_dir)
        print(f'force change working directory to {w_dir}')
    # save complex_pdbstr into w_dir
    complex_path = os.path.join(w_dir, 'complex.pdb')
    opts_file(complex_path, 'w', data=complex_pdbstr)
    # Run HBadd
    os.system(' '.join([cwd_cmd, os.path.join(ligplus_lib, "hbadd"), complex_path, os.path.join(ligplus_params, 'components.cif'), '-wkdir', w_dir]))
    os.system(' '.join([cwd_cmd, os.path.join(ligplus_lib, "hbplus"), '-L', '-h', '2.90', '-d', '3.90', '-N', complex_path, '-wkdir', w_dir]))
    os.system(' '.join([cwd_cmd, os.path.join(ligplus_lib, "hbplus"), '-L', '-h', '2.70', '-d', '3.35', complex_path, '-wkdir', w_dir]))
    os.system(' '.join([cwd_cmd, os.path.join(ligplus_lib, "dimer"), complex_path, receptor_chain, ligand_chain]))
    os.system(' '.join([cwd_cmd, os.path.join(ligplus_lib, "dimhtml"), 'none', '-dimp', '-dir', w_dir, '-flip', '-ctype', '1']))
    os.system(' '.join([cwd_cmd,
        os.path.join(ligplus_lib, "ligplot"), os.path.join(w_dir, 'dimplot.pdb'), '-wkdir', w_dir,
        '-prm', os.path.join(ligplus_params, 'dimplot.prm'), '-ctype', '1'
        ]))
    # parse output
    try:
        hhb_lines = parse_ligplus_output(os.path.join(w_dir, 'ligplot.hhb'))
        nnb_lines = parse_ligplus_output(os.path.join(w_dir, 'ligplot.nnb'))
    except FileNotFoundError:
        if kwargs.get('is_retry', False):
            print(f'LigPlus output file not found in {w_dir}, retry once')
            return _run_ligplus_for_complex(ligplus_dir, complex_pdbstr, receptor_chain, ligand_chain, mode, cutoff, is_retry=True)
        else:
            print(f'LigPlus output file not found in {w_dir}, set result to empty')
        hhb_lines, nnb_lines = [], []
    # try to close temporary directory
    if w_dir_handle is not None:
        try:
            w_dir_handle.cleanup()
        except Exception as e:
            print(f'Error when cleaning up temporary directory: {e}, skip')
    if force_cwd:
        os.chdir(now_wd)
        print(f'change working directory back to {now_wd}')
    hhb_lines = [(line[1], line[0], line[2]) if line[0][2] == ligand_chain else line for line in hhb_lines if line[2] <= cutoff]
    nnb_lines = [(line[1], line[0], line[2]) if line[0][2] == ligand_chain else line for line in nnb_lines if line[2] <= cutoff]
    interactions = {'Hydrogen Bonds': hhb_lines, 'Non-bonded Interactions': nnb_lines}
    return {k:v for k,v in interactions.items() if k in mode}

    
def run_ligplus(ligplus_dir: str, receptor: str = None, ligand: str = None,
                complex: str = None, receptor_chain: str = None, ligand_chain: str = None,
                mode: List[str] = None, cutoff: float = 4., taskpool: TaskPool = None,
                force_cwd: bool = False, w_dir: str = None, **kwargs) -> Dict[str, List[Tuple[Tuple[int, str, str], Tuple[int, str, str], float]]]:
    """
    run LigPlus to get atom-level interactions between receptor and ligand.
    Parameters:
        - ligplus_dir (str): LigPlus directory, which contains LigPlus.jar
        - receptor (str): receptor name in pymol, or pdb file path, default None
        - ligand (str): ligand name in pymol, or pdb file path, default None
        - complex (str): complex name in pymol, or pdb file path, default None
        - receptor_chain (str): receptor chain id, default None, only used when complex is provided
        - ligand_chain (str): ligand chain id, default None, only used when complex is provided
        - taskpool (TaskPool): TaskPool object for parallelization, default None
        - cutoff (float): cutoff distance, default 4.
    Returns:
        interactions (dict):
            - key: interaction type, includes: Hydrogen Bonds, Non-bonded Interactions
            - value: list of interaction info: ((receptor_resi, receptor_resn, receptor_chain), (ligand_resi, ligand_resn, ligand_chain), distance)
    """
    def load_mol(mol):
        if os.path.exists(mol) and os.path.isfile(mol):
            tmp_name = f'LAZYDOCK_TMP_OBJ_{uuid4()}'
            cmd.load(mol, tmp_name)
            chain = cmd.get_chains(tmp_name)[0]
            return tmp_name, chain
        elif mol in cmd.get_names():
            return mol, cmd.get_chains(mol)[0]
        else:
            put_err(f'{mol} not found in pymol or file path', _exit=True)
    # load receptor and ligand, make complex
    if complex is not None:
        if receptor_chain is None or ligand_chain is None:
            put_err('receptor_chain and ligand_chain must be provided when complex is provided', _exit=True)
        if (os.path.exists(complex) and os.path.isfile(complex)) or (complex in cmd.get_names()):
            complex_pdbstr = opts_file(complex_name)
        else:
            put_err(f'{complex} not found in pymol or file path', _exit=True)
    elif receptor is not None and ligand is not None:
        receptor_name, receptor_chain = load_mol(receptor)
        ligand_name, ligand_chain = load_mol(ligand)
        complex_name = f'LAZYDOCK_TMP_OBJ_{uuid4()}'
        cmd.select(complex_name, f'{receptor_name} or {ligand_name}')
        complex_pdbstr = cmd.get_pdbstr(complex_name)
        cmd.delete(complex_name) # this do not delete receptor and ligand
    else:
        put_err('either complex or receptor and ligand must be provided', _exit=True)
    # delete temporary objects if loaded from file
    for mol in [receptor_name, ligand_name]:
        if mol.startswith('LAZYDOCK_TMP_OBJ_'):
            cmd.delete(mol)
    # run LigPlus
    if taskpool is None:
        return _run_ligplus_for_complex(ligplus_dir, complex_pdbstr, receptor_chain, ligand_chain, mode, cutoff, force_cwd, w_dir)
    else:
        return taskpool.add_task(None, _run_ligplus_for_complex, ligplus_dir, complex_pdbstr, receptor_chain, ligand_chain, mode, cutoff, force_cwd, w_dir)


def merge_interaction_df(interaction: Dict[str, List[Tuple[Tuple[int, str, str], Tuple[int, str, str], float]]],
                         interaction_df: pd.DataFrame, distance_cutoff: float):
    """merge the interactions returned by calcu_atom_level_interactions to interaction_df."""
    # index format: CHAIN_ID:RESI:RESN
    for interaction_type, values in interaction.items():
        for single_inter in values:
            # single_inter: ((217, 'VAL', 'A'), (10, 'PHE', 'Z'), 3.71)
            receptor_res = f'{single_inter[0][2]}:{single_inter[0][0]}:{single_inter[0][1]}'
            ligand_res = f'{single_inter[1][2]}:{single_inter[1][0]}:{single_inter[1][1]}'
            points = distance_cutoff - single_inter[2]
            if ligand_res not in interaction_df.index or receptor_res not in interaction_df.columns:
                interaction_df.loc[ligand_res, receptor_res] = points
            elif np.isnan(interaction_df.loc[ligand_res, receptor_res]):
                interaction_df.loc[ligand_res, receptor_res] = points
            else:
                interaction_df.loc[ligand_res, receptor_res] += points
    return interaction_df


SUPPORTED_MODE = ['Hydrogen Bonds', 'Non-bonded Interactions']


def calcu_receptor_poses_interaction(receptor: str, poses: List[str], ligplus_dir: Optional[str] = None,
                                     mode: Union[str, List[str]] = 'Hydrogen Bonds', cutoff: float = 4.,
                                     taskpool: TaskPool = None, verbose: bool = False,
                                     force_cwd: bool = False, w_dir: str = None, **kwargs):
    """
    calcu interactions between one receptor and one ligand with many poses using LigPlus.
    Parameters:
        - receptor (str): receptor pymol name
        - poses (List[str]):  ligand pymol names
        - ligplus_dir (str): LigPlus directory, which contains LigPlus.jar
        - mode (str or List[str]): supported modes: Hydrogen Bonds, Non-bonded Interactions, default Hydrogen Bonds, can be set to 'all' to include all modes.
        - taskpool (TaskPool): TaskPool object for parallelization, default None
        - cutoff (float): cutoff distance, default 4
        
    Returns:
        interactions (dict):
            - key: ligand name
            - value: interactions dict between receptor and ligand
                - key: interaction type, includes: Hydrophobic Interactions, Hydrogen Bonds, Water Bridges, Salt Bridges, pi-Stacking, pi-Cation Interactions, Halogen Bonds, Metal Complexes
                - value: list of interaction info: ((receptor_resi, receptor_resn, receptor_chain), (ligand_resi, ligand_resn, ligand_chain), distance)
        
        interaction_df (pd.DataFrame): , interactions between receptor and ligand, in the format of ligand-residue-residue matrix, with the value of each cell is the interaction score between two atoms.
            interaction_df.loc[ligand_res, receptor_res] = score
    """
    # set mode
    if mode == 'all':
        mode = SUPPORTED_MODE
    # set ligplus_dir
    if ligplus_dir is None:
        ligplus_dir = config.GlobalConfig.named_paths['ligplus_dir']
        if ligplus_dir is None:
            put_err('LigPlus directory not provided and not found in config.json', _exit=True)
    # prepare interactions
    all_interactions, interaction_df = {}, pd.DataFrame()
    # calcu for each ligand
    for ligand in tqdm(poses, desc=f'LigPlus', disable=not verbose):
        # calcu interaction
        all_interactions[ligand] = run_ligplus(ligplus_dir, receptor, ligand, mode=mode, cutoff=cutoff, taskpool=taskpool,
                                               force_cwd=force_cwd, w_dir=w_dir)
        # wait for taskpool
        while taskpool is not None and taskpool.count_waiting_tasks() > 0:
            time.sleep(0.1)
    # merge interactions by res
    for ligand in all_interactions:
        if taskpool is not None:
            all_interactions[ligand] = taskpool.query_task(all_interactions[ligand], True, 10**8)
        # merge interactions by res
        merge_interaction_df(all_interactions[ligand], interaction_df, cutoff)
    if not interaction_df.empty:
        # sort res
        interaction_df.sort_index(axis=0, inplace=True, key=sort_func)
        interaction_df.sort_index(axis=1, inplace=True, key=sort_func)
        interaction_df.fillna(0, inplace=True)
    else:
        return None, None
    return all_interactions, interaction_df


if __name__ == '__main__':
    # dev code
    from lazydock.pml.autodock_utils import DlgFile
    cmd.reinitialize()
    cmd.load('data_tmp/pdb/RECEPTOR.pdb', 'RECEPTOR')
    dlg = DlgFile(path='data_tmp/dlg/1000run.dlg', sort_pdb_line_by_res=True, parse2std=True)
    dlg.sort_pose()
    pose_lst = []
    for i, pose in enumerate(dlg.pose_lst[:10]):
        pose_lst.append(f'ligand_{i}')
        cmd.read_pdbstr(pose.as_pdb_string(), pose_lst[-1])
        cmd.alter(f'ligand_{i}', 'type="HETATM"')
    result = run_ligplus(config.GlobalConfig.named_paths['ligplus_dir'],
                         receptor='RECEPTOR', ligand=pose_lst[0], mode='Hydrogen Bonds')
    taskpool = TaskPool('threads', n_worker=4).start()
    interactions, interaction_df = calcu_receptor_poses_interaction('RECEPTOR', pose_lst, cutoff=4, taskpool=None, w_dir='data_tmp/dlg')
    print(interactions)


'''
Date: 2024-12-18 10:48:32
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-12-26 09:53:33
Description:
'''
import os
from typing import Any, Dict, List

from mbapy_lite.base import get_fmt_time, put_log
from mbapy_lite.file import opts_file
from mbapy_lite.game import BaseInfo

if __name__ == '__main__':
    from lazydock.utils import get_storage_path
else:
    from ..utils import get_storage_path


class Gromacs(BaseInfo):
    def __init__(self, call_name: str = 'gmx', working_dir: str = '.'):
        super().__init__()
        self.call_name = call_name
        self.working_dir = os.path.abspath(working_dir)
        
    def kwargs2cmd(self, kwargs: Dict[str, str]):
        return ' '.join([f'{"--" if k.startswith("_") else "-"}{k} {v}' for k, v in kwargs.items()])
    
    def gen_command(self, sub_commmand: str, **kwargs):
        """
        Run gromacs command.
        
        Parameters:
            - sub_commmand: str, sub command of gromacs, such as grompp, mdrun, etc.
            - **kwargs: dict, keyword arguments for gromacs command.
                - if the key starts with '_', it will be added with '--' before the key, else with '-' before the key.
                - if the value is bool, it will be added with no value.
                - if the value is list or tuple, it will be joined with space.
                - if the value is not str, it will be converted to str directly via str().
        """
        for k in list(kwargs.keys()):
            if kwargs[k] is None:
                del kwargs[k]
            elif isinstance(kwargs[k], bool):
                if kwargs[k]:
                    kwargs[k] = ''
                else:
                    kwargs[f'no{k}'] = ''
            elif isinstance(kwargs[k], (list, tuple)):
                kwargs[k] =' '.join(map(str, kwargs[k]))
            elif not isinstance(kwargs[k], str):
                kwargs[k] = str(kwargs[k])
        return f'cd "{self.working_dir}" && {self.call_name} {sub_commmand} {self.kwargs2cmd(kwargs)}'
    
    def run_command_with_expect(self, sub_commmand: str, expect_actions: List[Dict[str, str]] = None,
                                expect_settings: Dict[str, Any] = None, enable_log: bool = False, **kwargs):
        """
        Run gromacs command with expect script.
        
        Parameters: 
            - sub_commmand: str, sub command of gromacs, such as grompp, mdrun, etc.
            - expect_actions: List[Dict[str, str]], the expect script actions.
                - key: the string to match the output of the command. if key is '\\timeout', it will be treated as a timeout.
                - value: the value to send
            - expect_settings: dict, the expect script settings.
                - timeout: int, default is -1, the timeout for expect script to start.
            - enable_log: bool, default is False, whether to enable log file.
            - **kwargs: dict, keyword arguments for gromacs command, generate by gen_command() method.
        """
        # set name and paths
        scripts_dir = os.path.join(self.working_dir, 'LazyDock_gmx_scripts')
        os.makedirs(scripts_dir, exist_ok=True)
        scripts_name = f'{get_fmt_time("%Y-%m-%d-%H-%M-%S.%f")}'
        # generate command
        cmd = self.gen_command(sub_commmand, **kwargs)
        if enable_log:
            log_path = os.path.join(scripts_dir, f'{scripts_name}.log')
            cmd = f'{cmd} > ./LazyDock_gmx_scripts/{scripts_name}.log'.replace('&&', '\n')
        put_log(f'Get command:\n{cmd}', head='LazyDock')
        # just run the command if no expect actions
        if expect_actions is None or not expect_actions:
            ret_val = os.system(cmd)
            if enable_log:
                ret_val = (ret_val, log_path)
            return ret_val
        # save cmd to bash file
        bash_path = os.path.join(scripts_dir, f'{scripts_name}.sh')
        opts_file(bash_path, 'w', data=cmd)
        # create expect script
        expect_settings = expect_settings or {}
        expect_lines = []
        expect_lines.append(f'set timeout {expect_settings.get("start_timeout", -1)}')
        expect_lines.append(f'spawn bash {bash_path}')
        for action in expect_actions:
            expect_lines.append('expect {')
            for key, value in action.items():
                if key == '\\timeout':
                    expect_lines.append(f'    timeout {{\n        puts "===TIMEOUT==="\n        send "{value}"}}')
                else:
                    expect_lines.append(f'    "{key}" {{\n        send "{value}"}}')
            expect_lines.append('}\n')
        expect_lines.append('interact')
        expect_script = '\n'.join(expect_lines)
        # save expect script to file and run it
        script_path = os.path.join(scripts_dir, f'{scripts_name}.exp')
        opts_file(script_path, 'w', data=expect_script)
        put_log(f'Running expect script: {script_path}', head='LazyDock')
        ret_val = os.system(f'cd "{self.working_dir}" && expect "{script_path}"')
        if enable_log:
            ret_val = (ret_val, log_path)
        return ret_val
        
        
if __name__ == '__main__':
    gmx = Gromacs()
    gmx.run_command_with_expect('grompp', f='topol.top', c='conf.gro', p='topol.top', o='tpr', maxwarn=1)
    gmx.run_command_with_expect('pdb2gmx -f receptor.pdb -o processed.gro -ter -ignh', [{')': '1\r'}, {'None': '1\r'}, {'None': '1\r'}, {'None': '1\r'}])


'''
Date: 2024-12-13 20:18:59
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-12-24 21:28:39
Description: steps most from http://www.mdtutorials.com/gmx
'''

import argparse
import os
import shutil
from pathlib import Path
from typing import Dict, List, Tuple, Union

from mbapy_lite.base import Configs, put_err, put_log
from mbapy_lite.file import get_paths_with_extension, opts_file
from mbapy_lite.web import Browser, TaskPool, random_sleep
from pymol import cmd
from tqdm import tqdm

from lazydock.config import CONFIG_FILE_PATH, GlobalConfig
from lazydock.gmx.run import Gromacs
from lazydock.gmx.thirdparty.cgenff_charmm2gmx import run_transform
from lazydock.gmx.thirdparty.sort_mol2_bonds import sort_bonds
from lazydock.pml.align_to_axis import align_pose_to_axis
from lazydock.pml.utils import get_seq
from lazydock.scripts._script_utils_ import Command, clean_path
from lazydock.web.cgenff import get_login_browser, get_result_from_CGenFF


class protein(Command):
    HELP = """
    prepare single protein for GROMACS MDS.
    
    STEPS:
    0. center complex.pdb by obabel, align with xyz axes by lazydock.
    1. prepare protein topology.
    """
    def __init__(self, args, printf = print):
        super().__init__(args, printf)
        
    @staticmethod
    def make_args(args: argparse.ArgumentParser):
        args.add_argument('-d', '--dir', type = str, default='.',
                          help='protein directory. Default is %(default)s.')
        args.add_argument('-n', '--protein-name', type = str,
                          help='protein file name in each sub-directory.')
        args.add_argument('--ff-dir', type = str,
                          help='force field files directory.')
        args.add_argument('--n-term', type = str, default='auto',
                          help='N-Term type for gmx pdb2gmx. Default is %(default)s.')
        args.add_argument('--pdb2gmx-args', type = str, default="-ter -ignh",
                          help='args pass to pdb2gmx command, default is %(default)s.')
        return args
    
    def process_args(self):
        self.args.dir = clean_path(self.args.dir)
        
    def main_process(self):
        # get protein paths
        if os.path.isdir(self.args.dir):
            proteins_path = get_paths_with_extension(self.args.dir, [], name_substr=self.args.protein_name)
        else:
            put_err(f'dir argument should be a directory: {self.args.config}, exit.', _exit=True)
        put_log(f'get {len(proteins_path)} protein(s)')
        # process each complex
        for protein_path in tqdm(proteins_path, total=len(proteins_path)):
            protein_path = Path(protein_path).resolve()
            gmx = Gromacs(working_dir=str(protein_path.parent))
            # STEP 0.1: center complex.pdb by obabel.
            ipath, opath = str(protein_path), str(protein_path.parent / f'0.1_{protein_path.stem}_center.pdb')
            if not os.path.exists(opath):
                os.system(f'obabel -ipdb {ipath} -opdb -O {opath} -c')
            # step 0.2: align complex.pdb with xyz axes by lazydock.
            ipath, opath = opath, str(protein_path.parent / f'0.2_{protein_path.stem}_center_align_axis.pdb')
            if not os.path.exists(opath):
                cmd.load(ipath, 'protein')
                align_pose_to_axis('protein')
                cmd.save(opath, 'protein')
                cmd.reinitialize()
            # STEP 1: Prepare the Protein Topology
            ipath, opath_rgro = opath, str(protein_path.parent / f'{protein_path.stem}.gro')
            if not os.path.exists(opath_rgro):
                if self.args.n_term == 'auto':
                    receptor_n_term = '1' if get_seq(ipath, fasta=False)[0] == 'P' else '0'
                    if receptor_n_term == '1':
                        put_log(f'using NH2 as N ternimal because the first residue of receptor is PRO.')
                else:
                    receptor_n_term = self.args.n_term
                gmx.run_command_with_expect(f'pdb2gmx -f {Path(ipath).name} -o {Path(opath_rgro).name} {self.args.pdb2gmx_args}',
                                            [{'dihedrals)': '1\r'}, {'None': '1\r'}, {'None': f'{receptor_n_term}\r'}, {'None': '0\r'}])


class ligand(Command):
    HELP = """
    prepare single ligand for GROMACS MDS.
    
    STEPS:
    0. center ligand.pdb by obabel
    1. alter chain code to "Z", align with xyz axes by lazydock.
    2. transfer ligand.pdb to mol2 by obabel.
    3. fix ligand name in mol2 file.
    4. sort mol2 bonds by lazydock.
    5. retrive str file from CGenFF.
    6. transfer str file to top and gro file by cgenff_charmm2gmx.py
    7. prepare ligand topology.
    """
    def __init__(self, args, printf = print):
        super().__init__(args, printf)
        
    @staticmethod
    def make_args(args: argparse.ArgumentParser):
        args.add_argument('-d', '--dir', type = str, default='.',
                          help='protein directory. Default is %(default)s.')
        args.add_argument('-n', '--ligand-name', type = str,
                          help='protein file name in each sub-directory.')
        args.add_argument('--ff-dir', type = str,
                          help='force field files directory, if before CGenFF step, \
can be a asbpath to charmmFF dir; \
if after CGenFF step, should be charmmFF dir name in each sub-directory, \
the program will use the ff-dir in sub-directory.')
        args.add_argument('--max-step', type = int, default=7,
                          help='max step to do. Default is %(default)s.')
        args.add_argument('--disable-browser', action='store_true',
                          help='whether to disable browser for CGenFF.')
        return args

    def process_args(self):
        self.args.dir = clean_path(self.args.dir)

    @staticmethod
    def insert_content(content: str, before: str, new_content: str):
        is_file, path = False, None
        if os.path.isfile(content):
            is_file, path = True, content
            content = opts_file(content)
        idx1 = content.find(before)
        if idx1 == -1:
            return put_err(f'{before} not found, skip.')
        content = content[:idx1+len(before)] + new_content + content[idx1+len(before):]
        if is_file:
            opts_file(path, 'w', data=content)
        return content

    @staticmethod
    def fix_name_in_mol2(ipath: str, opath: str):
        lines = opts_file(ipath, way='lines')
        lines[1] = 'LIG\n'
        get_idx_fn = lambda content: lines.index(list(filter(lambda x: x.startswith(content), lines))[0])
        atom_st, atom_ed = get_idx_fn('@<TRIPOS>ATOM')+1, get_idx_fn('@<TRIPOS>BOND')
        for i in range(atom_st, atom_ed):
            resn = lines[i].split()[7].strip()
            resn_idx = lines[i].index(resn)
            lines[i] = lines[i][:resn_idx] + 'LIG' + ' '*(min(0, len(resn) - 3)) + lines[i][resn_idx+len(resn):]
        opts_file(opath, 'w', way='lines', data=lines)

    @staticmethod
    def get_login_browser(download_dir: str):
        put_log(f'getting CGenFF account from {CONFIG_FILE_PATH}')
        email, password = GlobalConfig.named_accounts['CGenFF']['email'], GlobalConfig.named_accounts['CGenFF']['password']
        if email is None or password is None:
            return put_err('CGenFF email or password not found in config file, skip.')
        return get_login_browser(email, password, download_dir=download_dir)

    @staticmethod
    def get_str_from_CGenFF(mol2_path: str, zip_path: str, browser: Browser) -> Union[str, None]:
        put_log(f'getting str file from CGenFF for {mol2_path}')
        get_result_from_CGenFF(mol2_path, b=browser)
        download_path = Path(browser.download_path) / Path(mol2_path).with_suffix('.zip').name
        shutil.move(str(download_path), zip_path)
        
    def prepare_ligand(self, ligand_path: str, main_path: Path):
        # STEP 2: transfer ligand.pdb to mol2 by obabel.
        ipath, opath = ligand_path, str(main_path.parent / f'2_{main_path.stem}_ligand.mol2')
        if self.args.max_step >= 2 and (not os.path.exists(opath)):
            os.system(f'obabel -ipdb {ipath} -omol2 -O {opath}')
        # STEP 3: fix ligand name and residue name in mol2 file.
        ipath, opath = opath, str(main_path.parent / f'3_{main_path.stem}_ligand_named.mol2')
        if self.args.max_step >= 3 and (not os.path.exists(opath)):
            self.fix_name_in_mol2(ipath, opath)
        # STEP 4: sort mol2 bonds by lazydock.
        ipath, opath = opath, str(main_path.parent / f'4_{main_path.stem}_ligand_sorted.mol2')
        if self.args.max_step >= 4 and (not os.path.exists(opath)):
            opts_file(opath, 'w', data=sort_bonds(ipath))
        # STEP 5: retrive str file from CGenFF.
        ipath, opath_str, opath_mol2 = opath, str(main_path.parent / f'5_{main_path.stem}_ligand_sorted.str'), str(main_path.parent / f'5_{main_path.stem}_ligand_sorted.mol2')
        cgenff_path = Path(ipath).with_suffix('.zip')
        if self.args.max_step >= 5 and (not os.path.exists(cgenff_path)):
            self.get_str_from_CGenFF(ipath, cgenff_path, browser=self.browser)
        if self.args.max_step >= 5 and (not os.path.exists(opath_str) or not os.path.exists(opath_mol2)):
            for file_name, content in opts_file(cgenff_path, 'r', way='zip').items():
                opts_file(cgenff_path.parent / file_name.replace('4_', '5_'), 'wb', data=content)
        # STEP 6: transfer str file to top and gro file by cgenff_charmm2gmx.py
        ipath_str, ipath_mol2, opath_itp = opath_str, opath_mol2, str(main_path.parent / f'lig.itp')
        # check and copy ff-dir
        ## if ff-dir already in each sub-directory, do not overwrite it.
        if os.path.exists(main_path.parent / self.args.ff_dir):
            ff_dir = main_path.parent / self.args.ff_dir
            put_log(f'ff-dir already exists in {ff_dir}, skip.')
        ## if ff-dir is a asbpath to charmmFF dir, copy it to sub-directory.
        elif os.path.exists(self.args.ff_dir):
            ff_dir = main_path.parent / Path(self.args.ff_dir).name
            if self.args.max_step >= 6 and (not ff_dir.exists()):
                shutil.copytree(os.path.abspath(self.args.ff_dir), ff_dir, dirs_exist_ok=True)
        else:
            ff_dir = None
            put_err(f'cannot find ff_dir: {self.args.ff_dir} in {main_path.parent}, set ff_dir to None, skip.')
        if self.args.max_step >= 6 and (not os.path.exists(opath_itp)):
            run_transform('LIG', ipath_mol2, ipath_str, str(ff_dir))

    def main_process(self):
        # allocate browser for CGenFF
        if not self.args.disable_browser:
            self.browser = self.get_login_browser(str(self.args.dir))
        # get ligand paths
        if os.path.isdir(self.args.dir):
            ligands_path = get_paths_with_extension(self.args.dir, [], name_substr=self.args.ligand_name)
        else:
            put_err(f'dir argument should be a directory: {self.args.config}, exit.', _exit=True)
        put_log(f'get {len(ligands_path)} ligand(s)')
        # process each ligand
        for ligand_path in tqdm(ligands_path, total=len(ligands_path)):
            ligand_path = Path(ligand_path).resolve()
            gmx = Gromacs(working_dir=str(ligand_path.parent))
            cmd.reinitialize()
            # STEP 0: center complex.pdb by obabel.
            ipath, opath = str(ligand_path), str(ligand_path.parent / f'0_{ligand_path.stem}_center.pdb')
            if not os.path.exists(opath):
                os.system(f'obabel -ipdb {ipath} -opdb -O {opath} -c')
            # step 1: align complex.pdb with xyz axes by lazydock.
            ipath, opath_l = opath, str(ligand_path.parent / f'1_{ligand_path.stem}_center_align_axis.pdb')
            if not os.path.exists(opath_l):
                cmd.load(ipath, 'ligand')
                cmd.alter('ligand', 'chain="Z"')
                align_pose_to_axis('ligand')
                cmd.save(opath_l, 'ligand')
                cmd.reinitialize()
            # STEP 2 ~ 6: prepare ligand topology.
            self.prepare_ligand(opath_l, ligand_path)
            # STEP 7: Prepare the ligand Topology
            ipath, opath_gro = str(ligand_path.parent / f'lig_ini.pdb'), str(ligand_path.parent / f'lig.gro')
            if self.args.max_step >= 7 and (not os.path.exists(opath_gro)):
                cmd.load(ipath, 'ligand')
                cmd.alter('ligand', 'chain="Z"')
                cmd.save(ipath, 'ligand')
                # because may do not have Gromacs installed, so just try
                try:
                    gmx.run_command_with_expect(f'editconf -f lig_ini.pdb -o lig.gro')
                except Exception as e:
                    put_err(f'pdb2gmx failed: {e}, skip.')


class complex(ligand):
    HELP = """
    prepare complex for GROMACS MDS.
    - input complex.pdb should have two chains, one for receptor and one for ligand.
    - complex.pdb should already add hydrogens by Avogadro or other software.
    - complex.pdb is supposed to be aligned with the axes to save space when MDS.
    
    STEPS:
    0. center complex.pdb by obabel, align with xyz axes by lazydock.
    1. extract receptor and ligand from complex.pdb.
    2. transfer ligand.pdb to mol2 by obabel.
    3. fix ligand name in mol2 file.
    4. sort mol2 bonds by lazydock.
    5. retrive str file from CGenFF.
    6. transfer str file to top and gro file by cgenff_charmm2gmx.py
    7. prepare protein topology.
    8. prepare ligand topology.
    9. merge receptor and ligand gro into complex.gro, prepare topol.top.
    """
    def __init__(self, args, printf = print):
        super().__init__(args, printf)
        self.browser = None
        
    @staticmethod
    def make_args(args: argparse.ArgumentParser):
        args.add_argument('-d', '--dir', type = str, default='.',
                          help='complex directory. Default is %(default)s.')
        args.add_argument('-n', '--complex-name', type = str,
                          help='complex name in each sub-directory.')
        args.add_argument('--max-step', type = int, default=9,
                          help='max step to do. Default is %(default)s.')
        args.add_argument('--receptor-chain-name', type = str,
                          help='receptor chain name.')
        args.add_argument('--ligand-chain-name', type = str,
                          help='ligand chain name.')
        args.add_argument('--ff-dir', type = str,
                          help='force field files directory.')
        args.add_argument('--disable-browser', action='store_true',
                          help='whether to disable browser for CGenFF.')
        return args
    
    def process_args(self):
        self.args.dir = clean_path(self.args.dir)
        if self.args.max_step < 1:
            return put_err('max step should be greater or equal to 1.', _exit=True)
        
    @staticmethod
    def extract_receptor_ligand(ipath: str, receptor_chain_name: str, ligand_chain_name: str, opath_r: str, opath_l: str):
        cmd.load(ipath, 'complex')
        for mol, opath, chain in zip(['receptor', 'ligand'], [opath_r, opath_l], [receptor_chain_name, ligand_chain_name]):
            if cmd.select(mol, f'complex and chain {chain}') == 0:
                put_err(f'{mol} chain {chain} has zero atom in {ipath}, skip this complex.')
                return False
            else:
                cmd.save(opath, mol)
        return True

    @staticmethod
    def prepare_complex_topol(ipath_rgro: str, ipath_lgro: str, ipath_top: str, opath_cgro: str, opath_top: str):
        # merge receptor and ligand gro into complex.gro
        receptor_gro_lines = list(filter(lambda x: len(x.strip()), opts_file(ipath_rgro, 'r', way='lines')))
        lig_gro_lines = list(filter(lambda x: len(x.strip()), opts_file(ipath_lgro, 'r', way='lines')))
        complex_gro_lines = receptor_gro_lines[:-1] + lig_gro_lines[2:-1] + receptor_gro_lines[-1:]
        complex_gro_lines[1] = f'{int(receptor_gro_lines[1]) + int(lig_gro_lines[1])}\n'
        opts_file(opath_cgro, 'w', way='lines', data=complex_gro_lines)
        # inset ligand paramters in topol.top
        topol = opts_file(ipath_top)
        topol = complex.insert_content(topol, '#include "posre.itp"\n#endif\n',
                                    '\n; Include ligand topology\n#include "lig.itp"\n')
        topol = complex.insert_content(topol, '#include "./charmm36-jul2022.ff/forcefield.itp"\n',
                                    '\n; Include ligand parameters\n#include "lig.prm"\n')
        topol += 'LIG                 1'
        opts_file(opath_top, 'w', data=topol)
        
    def main_process(self):
        # allocate browser for CGenFF
        if not self.args.disable_browser:
            self.browser = self.get_login_browser(str(self.args.dir))
        # get complex paths
        if os.path.isdir(self.args.dir):
            complexs_path = get_paths_with_extension(self.args.dir, [], name_substr=self.args.complex_name)
        else:
            put_err(f'dir argument should be a directory: {self.args.config}, exit.', _exit=True)
        put_log(f'get {len(complexs_path)} complex(s)')
        # process each complex
        for complex_path in tqdm(complexs_path, total=len(complexs_path)):
            complex_path = Path(complex_path).resolve()
            gmx = Gromacs(working_dir=str(complex_path.parent))
            cmd.reinitialize()
            # STEP 0.1: center complex.pdb by obabel.
            ipath, opath = str(complex_path), str(complex_path.parent / f'0.1_{complex_path.stem}_center.pdb')
            if not os.path.exists(opath):
                os.system(f'obabel -ipdb {ipath} -opdb -O {opath} -c')
            # step 0.2: align complex.pdb with xyz axes by lazydock.
            ipath, opath = opath, str(complex_path.parent / f'0.2_{complex_path.stem}_center_align_axis.pdb')
            if not os.path.exists(opath):
                cmd.load(ipath, 'complex')
                align_pose_to_axis('complex')
                cmd.save(opath, 'complex')
                cmd.reinitialize()
            # STEP 1: extract receptor and ligand from complex.pdb.
            ipath, opath_r, opath_l = opath, str(complex_path.parent / f'1_{complex_path.stem}_receptor.pdb'), str(complex_path.parent / f'1_{complex_path.stem}_ligand.pdb')
            if self.args.max_step >= 1 and (not os.path.exists(opath_r) or not os.path.exists(opath_l)):
                if not self.extract_receptor_ligand(ipath, self.args.receptor_chain_name, self.args.ligand_chain_name, opath_r, opath_l):
                    continue
            # STEP 2 ~ 6: prepare ligand topology.
            self.prepare_ligand(opath_l, complex_path)
            # STEP 7: Prepare the Protein Topology
            ipath, opath_rgro = opath_r, str(complex_path.parent / f'{complex_path.stem}_receptor.gro')
            if self.args.max_step >= 7 and (not os.path.exists(opath_rgro)):
                receptor_n_term = '1' if get_seq(opath_r, fasta=False)[0] == 'P' else '0'
                if receptor_n_term == '1':
                    put_log(f'using NH2 as N ternimal because the first residue of receptor is PRO.')
                gmx.run_command_with_expect(f'pdb2gmx -f {Path(ipath).name} -o {Path(opath_rgro).name} -ter',
                                            [{'dihedrals)': '1\r'}, {'None': '1\r'}, {'None': f'{receptor_n_term}\r'}, {'None': '0\r'}])
            # STEP 8: Prepare the Ligand Topology
            opath_lgro = str(complex_path.parent / 'lig.gro')
            if self.args.max_step >= 8 and (not os.path.exists(opath_lgro)):
                gmx.run_command_with_expect('editconf -f lig_ini.pdb -o lig.gro')
            # STEP 9: Prepare the Complex Topology
            opath_cgro, opath_top = str(complex_path.parent / 'complex.gro'), str(complex_path.parent / 'topol.top')
            if self.args.max_step >= 9 and (not os.path.exists(opath_cgro)):
                self.prepare_complex_topol(opath_rgro, opath_lgro, opath_top, opath_cgro, opath_top)


_str2func = {
    'protein': protein,
    'ligand': ligand,
    'complex': complex,
}

def main(sys_args: List[str] = None):
    args_paser = argparse.ArgumentParser(description = 'tools for GROMACS.')
    subparsers = args_paser.add_subparsers(title='subcommands', dest='sub_command')

    prepare_protein_args = protein.make_args(subparsers.add_parser('protein', description=protein.HELP))
    prepare_ligand_args = ligand.make_args(subparsers.add_parser('ligand', description=ligand.HELP))
    prepare_complex_args = complex.make_args(subparsers.add_parser('complex', description=complex.HELP))

    args = args_paser.parse_args(sys_args)
    if args.sub_command in _str2func:
        _str2func[args.sub_command](args).excute()


if __name__ == "__main__":
    # pass
    # main(r'complex -d data_tmp/gmx/complex -n complex.pdb --receptor-chain-name A --ligand-chain-name Z --ff-dir data_tmp/gmx/charmm36-jul2022.ff'.split())
    
    main()


import argparse
import os
from pathlib import Path
from typing import Dict, List, Tuple, Union

from mbapy_lite.base import put_err, put_log
from mbapy_lite.file import get_paths_with_extension, opts_file
from pymol import cmd
from tqdm import tqdm

from lazydock.gmx.run import Gromacs
from lazydock.scripts._script_utils_ import Command, clean_path
from lazydock.utils import uuid4


class simple(Command):
    HELP = """
    simple analysis for GROMACS simulation
    0. gmx_mpi trjconv -s md.tpr -f md.xtc -o md_center.xtc -pbc mol -center
    
    1. gmx_mpi rms -s md.tpr -f md_center.xtc -o rmsd.xvg -tu ns 
    2. gmx_mpi rmsf -s md.tpr -f md_center.xtc -o rmsf.xvg
    3. gmx_mpi gyrate -s md.tpr -f md_center.xtc -o gyrate.xvg
    4. gmx_mpi hbond -s md.tpr -f md_center.xtc -num -dt 10
    
    5. gmx_mpi sasa -s md.tpr -f md_center.xtc -o sasa_total.xvg -or sasa_res.xvg -tu ns 
    6. gmx_mpi covar -s md.tpr -f md_center.xtc -o eigenval.xvg -tu ns 
    """
    def __init__(self, args, printf=print):
        super().__init__(args, printf)
        
    @staticmethod
    def make_args(args: argparse.ArgumentParser):
        args.add_argument('-d', '--dir', type=str,
                          help='directory to store the prepared files')
        args.add_argument('-n', '--main-name', type = str,
                          help='main name in each sub-directory, such as md.tpr.')           
        return args
        
    @staticmethod
    def trjconv(gmx: Gromacs, main_name: str, center_group: str = '1', **kwargs):
        gmx.run_command_with_expect('trjconv', s=f'{main_name}.tpr', f=f'{main_name}.xtc', o=f'{main_name}_center.xtc', pbc='mol', center=True,
                                    expect_actions=[{'Select a group:': f'{center_group}\r'}, {'Select a group:': '0\r'}], **kwargs)
        
    @staticmethod
    def rms(gmx: Gromacs, main_name: str, group: str = '4', **kwargs):
        gmx.run_command_with_expect('rms', s=f'{main_name}.tpr', f=f'{main_name}_center.xtc', o=f'rmsd.xvg', tu='ns',
                                    expect_actions=[{'Select a group:': f'{group}\r'}, {'Select a group:': f'{group}\r'}], **kwargs)
        os.system(f'cd "{gmx.working_dir}" && dit xvg_show -f rmsd.xvg -o rmsd.png -smv')
        
    @staticmethod
    def rmsf(gmx: Gromacs, main_name: str, group: str = '4', res: bool = True, **kwargs):
        gmx.run_command_with_expect('rmsf', s=f'{main_name}.tpr', f=f'{main_name}_center.xtc', o=f'rmsf.xvg', res=res,
                                    expect_actions=[{'Select a group:': f'{group}\r'}], **kwargs)
        os.system(f'cd "{gmx.working_dir}" && dit xvg_show -f rmsf.xvg -o rmsf.png -smv')
        
    @staticmethod
    def gyrate(gmx: Gromacs, main_name: str, group: str = '4', **kwargs):
        gmx.run_command_with_expect('gyrate', s=f'{main_name}.tpr', f=f'{main_name}_center.xtc', o=f'gyrate.xvg',
                                    expect_actions=[{'Select a group:': f'{group}\r'}], **kwargs)
        os.system(f'cd "{gmx.working_dir}" && dit xvg_show -f gyrate.xvg -o gyrate.png -smv')
        
    @staticmethod
    def hbond(gmx: Gromacs, main_name: str, group: str = '1', dt=10, **kwargs):
        gmx.run_command_with_expect('hbond', s=f'{main_name}.tpr', f=f'{main_name}_center.xtc',
                                    num=f'{main_name}_hbond_num.xvg', dist=f'{main_name}_hbond_dist.xvg',
                                    expect_actions=[{'Select a group:': f'{group}\r'}, {'Select a group:': f'{group}\r'}], **kwargs)
        for ty in ['num', 'dist']:
            os.system(f'cd "{gmx.working_dir}" && dit xvg_show -f {main_name}_hbond_{ty}.xvg -o hbond_{ty}.png -smv')
        
    @staticmethod
    def sasa(gmx: Gromacs, main_name: str, group: str = '1', **kwargs):
        gmx.run_command_with_expect('sasa or sasa_res.xvg', s=f'{main_name}.tpr', f=f'{main_name}_center.xtc',
                                    o=f'sasa_total.xvg', odg=f'sasa_dg.xvg', tv='sasa_tv.xvg', tu='ns',
                                    expect_actions=[{'Select a group:': f'{group}\r'}], **kwargs)
        for ty in ['total', 'res', 'dg', 'tv']:
            os.system(f'cd "{gmx.working_dir}" && dit xvg_show -f sasa_{ty}.xvg -o sasa_{ty}.png -smv')
        
    @staticmethod
    def covar(gmx: Gromacs, main_name: str, xmax: str = 10, **kwargs):
        gmx.run_command_with_expect('covar', s=f'{main_name}.tpr', f=f'{main_name}_center.xtc', o=f'eigenval.xvg', tu='ns',
                                    expect_actions=[{'Select a group:': '1\r'}], **kwargs)
        os.system(f'cd "{gmx.working_dir}" && dit xvg_show -f eigenval.xvg -o eigenval.png -xmin 0 -xmax {xmax}')
        
    def main_process(self):
        # get complex paths
        if os.path.isdir(self.args.dir):
            complexs_path = get_paths_with_extension(self.args.dir, [], name_substr=self.args.main_name)
        else:
            put_err(f'dir argument should be a directory: {self.args.dir}, exit.', _exit=True)
        put_log(f'get {len(complexs_path)} task(s)')
        # process each complex
        for complex_path in tqdm(complexs_path, total=len(complexs_path)):
            complex_path = Path(complex_path).resolve()
            gmx = Gromacs(working_dir=str(complex_path.parent))
            # perform trjconv
            self.trjconv(gmx, main_name=complex_path.stem, center_group='1')
            # perform analysis
            self.rms(gmx, main_name=complex_path.stem, group='4')
            self.rmsf(gmx, main_name=complex_path.stem, group='4')
            self.gyrate(gmx, main_name=complex_path.stem, group='4')
            self.hbond(gmx, main_name=complex_path.stem, group='1')
            self.sasa(gmx, main_name=complex_path.stem, group='1')
            self.covar(gmx, main_name=complex_path.stem, xmax=10)


_str2func = {
    'simple': simple,
}


def main(sys_args: List[str] = None):
    args_paser = argparse.ArgumentParser(description = 'tools for GROMACS analysis.')
    subparsers = args_paser.add_subparsers(title='subcommands', dest='sub_command')

    simple_args = simple.make_args(subparsers.add_parser('simple', description=simple.HELP))

    args = args_paser.parse_args(sys_args)
    if args.sub_command in _str2func:
        _str2func[args.sub_command](args).excute()


if __name__ == '__main__':
    main()


'''
Date: 2024-12-07 21:04:01
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-12-08 11:08:58
Description: 
'''

import os
from pathlib import Path
from typing import Dict

import pandas as pd
import requests
from mbapy_lite.base import put_err, put_log
from mbapy_lite.file import opts_file
from mbapy_lite.web import Browser, random_sleep


def get_score_from_SwissADME(lig_SMILES: str, result_dir: str, browser: Browser = None, timeout: int = 1200, **kwargs) -> Dict[str, str]:
    """return Score in dict"""
    b = browser or Browser()
    b.browser.fullscreen_window()
    b.get('http://www.swissadme.ch/index.php')
    b.send_key(key=lig_SMILES, element='//*[@id="smiles"]')
    b.execute_script("arguments[0].removeAttribute('disabled')", b.find_elements(element='//*[@id="submitButton"]')[0])
    b.click(element='//*[@id="submitButton"]')
    xpath_result_img1 = '//*[@id="content"]/div[11]/div[1]/div[3]/div[2]/img'
    xpath_limit_err = '//*[@id="content"]'
    if not b.wait_element([xpath_result_img1, xpath_limit_err], timeout=timeout, check_fn=any):
        return put_err('Timeout', {})
    err_info = b.find_elements(xpath_limit_err)
    err_text = 'You are submitting molecules that are too large'
    if err_info and err_text in err_info[0].text:
        return put_err('Too large', {'ERROR': err_text})
    # download img1
    img1_url = b.find_elements(xpath_result_img1)[0].get_attribute('src')
    img1_path = os.path.join(result_dir, 'SwissADME_img1.png')
    opts_file(img1_path, 'wb', data=requests.get(img1_url).content)
    # cap img2
    b.click(element='//*[@id="content"]/div[8]/span[1]/button')
    img2 = b.find_elements('//*[@id="placeholder"]/canvas[2]')[0].screenshot_as_png
    img2_path = os.path.join(result_dir, 'SwissADME_img2.png')
    opts_file(img2_path, 'wb', data=img2)
    # download scv file
    csv_link = b.find_elements('//*[@id="content"]/div[7]/a[1]')[0].get_attribute('href')
    csv_path = os.path.join(result_dir,'result.csv')
    opts_file(csv_path, 'wb', data=requests.get(csv_link).content)
    # transpose csv file
    csvT_path = os.path.join(result_dir,'resultT.csv')
    pd.read_csv(csv_path).T.to_csv(csvT_path, index=True)
    return {'img1_path': img1_path, 'img2_path': img2_path, 'csv_path': csv_path, 'csvT_path': csvT_path}


def get_score_from_SwissTargetPrediction(lig_SMILES: str, result_dir: str, browser: Browser = None, timeout: int = 1200, **kwargs) -> Dict[str, str]:
    """return Score in dict"""
    result_dir = str(Path(result_dir).resolve())
    if not browser:
        b = Browser(download_path=result_dir)
        put_log(f'using download path: {result_dir}')
    else:
        b = browser
    b.browser.fullscreen_window()
    b.get('http://www.swisstargetprediction.ch/')
    b.send_key(key=lig_SMILES, element='//*[@id="smilesBox"]')
    b.execute_script("arguments[0].removeAttribute('disabled')", b.find_elements(element='//*[@id="submitButton"]')[0])
    b.click(element='//*[@id="submitButton"]')
    xpath_result_img1 = '//*[@id="content"]/div[3]/img'
    xpath_limit_err = '//*[@id="content"]'
    if not b.wait_element([xpath_result_img1, xpath_limit_err], timeout=timeout, check_fn=any):
        return put_err('Timeout', {})
    err_info = b.find_elements(xpath_limit_err)
    err_text = 'You are submitting molecules that are too large'
    if err_info and err_text in err_info[0].text:
        return put_err('Too large', {'ERROR': err_text})
    # download img1
    img1_url = b.find_elements(xpath_result_img1)[0].get_attribute('src')
    img1_path = os.path.join(result_dir, 'SwissTargetPrediction_img1.png')
    opts_file(img1_path, 'wb', data=requests.get(img1_url).content)
    # download img2
    img2_url = b.find_elements('//*[@id="placePieChart"]/a/img')[0].get_attribute('src')
    img2_path = os.path.join(result_dir, 'SwissTargetPrediction_img2.png')
    opts_file(img2_path, 'wb', data=requests.get(img2_url).content)
    # download csv file
    b.click(element='//*[@id="exportButtons"]/div/button[2]')
    b.click(element='//*[@id="exportButtons"]/div/button[3]')
    return {'img1_path': img1_path, 'img2_path': img2_path}


if __name__ == '__main__':
    from mbapy.base import Configs
    b = Browser(options = [f"--user-agent={Configs.web.chrome_driver_path}"], download_path='data_tmp/web')
    lig_SMILES = '[H]N[C@H](C(N1CCC[C@H]1C(N[C@H](C(N2CCC[C@H]2C(N[C@@H](C(C)C)C(N[C@@H](CC(N)=O)C(N[C@H](C(N[C@@H](CCCCN)C(N[C@@H](CC(C)C)C(N[C@@H](CC(C)C)C(N[C@@H](CO)C(N[C@H](C(N)=O)CC3=CN=CN3)=O)=O)=O)=O)=O)CC4=CC=CC=C4)=O)=O)=O)=O)CC5=CC=CC=C5)=O)=O)CC6=CC=C(O)C=C6'
    result_dir = 'data_tmp/web'
    get_score_from_SwissTargetPrediction(lig_SMILES, result_dir, browser=b, timeout=1200)


'''
Date: 2024-12-16 15:32:10
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-12-16 17:51:35
Description: 
'''
import argparse
import os
import time
from pathlib import Path
from typing import Dict, List

import pandas as pd
from tqdm import tqdm
from mbapy_lite.base import get_fmt_time, put_err
from mbapy_lite.file import (get_paths_with_extension, is_jsonable, opts_file,
                             write_sheets)
from mbapy_lite.web import TaskPool, random_sleep
from pymol import cmd

from lazydock.scripts._script_utils_ import clean_path, show_args
from lazydock.web.model_eval import get_eval_info_from_servers


def get_info(result_path, *args, **kwargs):
    result = get_eval_info_from_servers(*args, **kwargs)
    opts_file(result_path, 'wb', way='pkl', data=result)
    return result


def asign_chain_code(pdb_path: str, save_path: str, name: str, code: str):
    cmd.load(pdb_path, name)
    cmd.alter(f'{name}', f'chain="{code}"')
    cmd.save(save_path, name)
    cmd.delete(name)


def main(sys_args: List[str] = None):
    SUPPORT_SERVERS = ['ProQ', 'VoroMQA', 'ProSA', 'MolProbity', 'ProQ3', 'SAVES', 'QMEANDisCo', 'QMEAN']
    args_paser = argparse.ArgumentParser(description = 'Run model evaluation on web servers.')
    args_paser.add_argument('-d', '--dir', type = str, default='.',
                            help='input directory. Default is %(default)s.')
    args_paser.add_argument('-n', '--name', type = str, default='',
                            help='sub string of file name. Default is %(default)s.')
    args_paser.add_argument('-t', '--type', type = str, default='pdb',
                            help='file type to search in input directory. Default is %(default)s.')
    args_paser.add_argument('--servers', type=str, default=','.join(SUPPORT_SERVERS),
                            help=f'web servers to evaluate, split by comma. Default is {SUPPORT_SERVERS}.')
    args_paser.add_argument('--n-workers', type=int, default=4,
                            help='number of workers to parallel. Default is %(default)s.')
    args_paser.add_argument('-r', '--recursive', action='store_true', default=False,
                            help='FLAG, recursive search. Default is %(default)s.')
    args_paser.add_argument('--disable-cache', action='store_true', default=False,
                            help='FLAG, whether to disable cache. Default is %(default)s.')
    args = args_paser.parse_args(sys_args)
    # process IO path
    args.dir = clean_path(args.dir)
    if os.path.isdir(args.dir):
        paths = get_paths_with_extension(args.dir, [args.type], name_substr=args.name, recursive=args.recursive)
        paths = list(map(lambda p: Path(p).resolve(), paths))
    else:
        return put_err(f'Input dir {args.dir} is not a dir, exit.', _exit=True)
    # process servers
    args.servers = args.servers.split(',')
    if not all(s in SUPPORT_SERVERS for s in args.servers):
        unsupports = list(set(args.servers) - set(SUPPORT_SERVERS))
        return put_err(f'Unsupported servers: {unsupports}.', _exit=True)
    # show args
    show_args(args, ['dir', 'name', 'n_workers','recursive', 'disable_cache'])
    # run tasks
    cmd.reinitialize()
    tasks = []
    taskpool = TaskPool('threads', args.n_workers).start()
    for path in tqdm(paths, total=len(paths), desc='Running tasks'):
        chain_alter_path = path.parent / f'{path.stem}_chain_alter{path.suffix}'
        result_path = os.path.join(args.dir, f'{chain_alter_path}.pkl')
        if not args.disable_cache and os.path.exists(result_path):
            print(f'Skip {path}, result file exists: {result_path}.')
            tasks.append(result_path)
            continue
        while taskpool.count_waiting_tasks() > 0:
            time.sleep(5)
        asign_chain_code(path, chain_alter_path, path.stem, 'A')
        task_name = taskpool.add_task(result_path, get_info, # task name is the result path
                                      result_path, chain_alter_path,
                                      servers=args.servers)
        tasks.append(task_name)
    taskpool.wait_till_tasks_done(tasks)
    # gather results
    df = pd.DataFrame(columns=['path'])
    df.set_index(['path'], inplace=True)
    for result_path, path in zip(tasks, paths):
        data = opts_file(result_path, 'rb', way='pkl')
        for server, result in data.items():
            for name, value in result.items():
                if is_jsonable(value):
                    df.loc[path, f'{server}:{name}'] = value
    df.sort_index(inplace=True)
    df.to_excel(os.path.join(args.dir, f'eval_{get_fmt_time()}.xlsx'))


if __name__ == "__main__":
    main()


'''
Date: 2024-11-27 17:24:03
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-12-25 19:55:40
Description: 
'''
import argparse
import os
from pathlib import Path
from typing import Callable, Dict, List, Tuple

import pandas as pd
from mbapy_lite.base import put_err
from mbapy_lite.file import get_paths_with_extension
from pymol import cmd
from tqdm import tqdm

if __name__ == '__main__':
    from lazydock.pml.autodock_utils import DlgFile
    from lazydock.pml.interaction_utils import SUPPORTED_MODE as pml_mode
    from lazydock.pml.interaction_utils import \
        calcu_receptor_poses_interaction as calc_fn_pml
    from lazydock.pml.ligplus_interaction import SUPPORTED_MODE as ligplus_mode
    from lazydock.pml.ligplus_interaction import \
        calcu_receptor_poses_interaction as calc_fn_ligplus
    from lazydock.pml.plip_interaction import SUPPORTED_MODE as plip_mode
    from lazydock.pml.plip_interaction import \
        calcu_receptor_poses_interaction as calc_fn_plip
    from lazydock.scripts._script_utils_ import Command, clean_path
else:
    from ..pml.autodock_utils import DlgFile
    from ..pml.interaction_utils import SUPPORTED_MODE as pml_mode
    from ..pml.interaction_utils import \
        calcu_receptor_poses_interaction as calc_fn_pml
    from ..pml.ligplus_interaction import SUPPORTED_MODE as ligplus_mode
    from ..pml.ligplus_interaction import \
        calcu_receptor_poses_interaction as calc_fn_ligplus
    from ..pml.plip_interaction import SUPPORTED_MODE as plip_mode
    from ..pml.plip_interaction import \
        calcu_receptor_poses_interaction as calc_fn_plip
    from ._script_utils_ import Command, clean_path
    
    
class simple_analysis(Command):
    METHODS: Dict[str, Tuple[Callable, List[str]]] = {'pymol': (calc_fn_pml, pml_mode),
                                                      'ligplus': (calc_fn_ligplus, ligplus_mode),
                                                      'plip': (calc_fn_plip, plip_mode)}
    def __init__(self, args: argparse.Namespace, printf=print) -> None:
        super().__init__(args, printf)
        self.tasks = []
        
    @staticmethod
    def make_args(args: argparse.ArgumentParser):
        args.add_argument('-r', '--receptor', type = str,
                          help="receptor pdb file path or file name, will be loaded by pymol.")
        args.add_argument('-l', '--ligand', type = str,
                          help=f"docking result path or file name, support Vina(pdbqt) and AutoDock(dlg) format.")
        args.add_argument('-bd', '--batch-dir', type = str, default=None,
                          help=f"dir which contains many sub-folders, each sub-folder contains docking result files.")
        args.add_argument('--method', type = str, default='pymol', choices=simple_analysis.METHODS.keys(),
                          help='search input directory recursively, default is %(default)s.')
        args.add_argument('--mode', type = str, default='all',
                          help=f'interaction mode, multple modes can be separated by comma, all method support `\'all\'` model.\npymol: {",".join(pml_mode)}\nligplus: {",".join(ligplus_mode)}\nplip: {",".join(plip_mode)}')
        args.add_argument('--cutoff', type = float, default=4,
                          help='distance cutoff for interaction calculation, default is %(default)s.')
        args.add_argument('--hydrogen-atom-only', default=False, action='store_true',
                          help='only consider hydrogen bond acceptor and donor atoms, this only works when method is pymol, default is %(default)s.')
        args.add_argument('--output-style', type = str, default='receptor', choices=['receptor'],
                          help='output style\n receptor: resn resi distance')
        return args

    def process_args(self):
        # process IO
        if self.args.batch_dir:
            self.args.batch_dir = clean_path(self.args.batch_dir)
        else:
            def check_file(path: str, name: str):
                path = clean_path(path)
                if not os.path.isfile(path):
                    put_err(f"{name} file is not a file path: {path}, exit.")
                    exit(1)
                if not os.path.exists(path):
                    put_err(f"{name} file is not exists: {path}, exit.")
                    exit(1)
                return path
            self.args.receptor = check_file(self.args.receptor, "receptor")
            self.args.ligand = check_file(self.args.ligand, "ligand")
        # check method and mode
        if ',' in self.args.mode:
            self.args.mode = [m.strip() for m in self.args.mode.split(',')]
        all_modes = set(simple_analysis.METHODS[self.args.method][1] + ['all'])
        if isinstance(self.args.mode, str) and self.args.mode not in all_modes:
            put_err(f"Unsupported mode: {self.args.mode}, supported mode: {simple_analysis.METHODS[self.args.method][1]}, exit.", _exit=True)
        elif isinstance(self.args.mode, list) and any(m not in all_modes for m in self.args.mode):
            unsuuported_mode = [m for m in self.args.mode if m not in all_modes]
            put_err(f"the mode you input has unsupported item(s): {unsuuported_mode}, supported mode: {simple_analysis.METHODS[self.args.method][1]}, exit.", _exit=True)
        
    @staticmethod
    def output_fromater_receptor(inter_value: Dict[str, float], method: str):
        # pymol: [('receptor', '', 'GLY', '300', 'O', 2817), ('LIGAND_0', 'Z', 'UNK', '1', 'N', 74), 2.8066137153155943]
        if method == 'pymol':
            inter_value = sorted(inter_value, key=lambda x: int(x[0][3]))
            return '; '.join(f'{v[0][2]}{v[0][3]}-{v[2]:.2f}' for v in inter_value)
        elif method in {'ligplus', 'plip'}:
            inter_value = sorted(inter_value, key=lambda x: int(x[0][0]))
            return '; '.join(f'{v[0][1]}{v[0][0]}-{v[2]:.2f}' for v in inter_value)
        else:
            put_err(f"Unsupported method: {method}, exit.")
            exit(1)
            
    @staticmethod
    def calc_interaction_from_dlg(receptor_path: str, dlg_path: str, method: str, mode: List[str], cutoff: float,
                                  output_formater: Callable, hydrogen_atom_only: bool = True) -> None:
        bar = tqdm(desc='Calculating interaction', leave=False)
        root = os.path.abspath(os.path.dirname(dlg_path))
        w_dir = os.path.join(root, 'ligplus') if method == 'ligplus' else root
        # load receptor
        receptor_name = Path(receptor_path).stem
        cmd.load(receptor_path, receptor_name)
        cmd.alter(receptor_name, 'chain="A"')
        bar.set_description(f'receptor loaded from {receptor_path}')
        # load poses from dlg and perform analysis
        # load poses
        dlg = DlgFile(dlg_path, None, True, True)
        bar.set_description(f'dlg loaded from {dlg_path}')
        dlg.sort_pose() # sort by docking energy
        pose_names = []
        for i, pose in enumerate(dlg.pose_lst):
            pose_names.append(f'LIGAND_{i}')
            cmd.read_pdbstr(pose.as_pdb_string(), pose_names[-1])
            cmd.alter(pose_names[-1], 'chain="Z"')
            cmd.alter(pose_names[-1], 'type="HETATM"')
        bar.set_description(f'{len(pose_names)} pose loaded')
        # calcu interactions
        fn, _ = simple_analysis.METHODS[method]
        bar.set_description(f'performing {method} calculation')
        interactions, _ = fn(receptor_name, pose_names, mode=mode, cutoff=cutoff, verbose=True, force_cwd=True, w_dir=w_dir, hydrogen_atom_only=hydrogen_atom_only)
        if interactions is None:
            cmd.reinitialize()
            return put_err(f"No interactions found in {dlg_path}")
        if method == 'pymol':
            interactions = {k:v[-1] for k,v in interactions.items()}
        bar.set_description(f'{method} interactions calculated')
        # save interactions
        df = pd.DataFrame()
        for i, (pose, interaction) in enumerate(zip(dlg.pose_lst, interactions.values())):
            df.loc[i, 'energy'] = pose.energy
            for inter_mode, inter_value in interaction.items():
                df.loc[i, inter_mode] = output_formater(inter_value, method)
        df.to_excel(os.path.join(root, f'{Path(dlg_path).stem}_{method}_interactions.xlsx'))
        bar.set_description(f'{method} interactions saved')
        # release all in pymol
        cmd.reinitialize()

    def main_process(self):
        # load origin dfs from data file
        if self.args.batch_dir:
            r_paths = get_paths_with_extension(self.args.batch_dir, ['.pdb', '.pdbqt'], name_substr=self.args.receptor)
            l_paths = get_paths_with_extension(self.args.batch_dir, ['.pdbqt', '.dlg'], name_substr=self.args.ligand)
            if len(r_paths) != len(l_paths):
                r_roots = [os.path.dirname(p) for p in r_paths]
                l_roots = [os.path.dirname(p) for p in l_paths]
                roots_count = {root: r_roots.count(root)+l_roots.count(root) for root in (set(r_roots) | set(l_roots))}
                invalid_roots = '\n'.join([root for root, count in roots_count.items() if count != 2])
                return put_err(f"The number of receptor and ligand files is not equal, please check the input files.\ninvalid roots:{invalid_roots}")
            for r_path, l_path in zip(r_paths, l_paths):
                self.tasks.append((r_path, l_path, self.args.method, self.args.mode))
        else:
            self.tasks.append((self.args.receptor, self.args.ligand, self.args.method, self.args.mode))
        # run tasks
        print(f'found {len(self.tasks)} tasks.')
        bar = tqdm(total=len(self.tasks), desc='Calculating interaction')
        for r_path, l_path, method, mode in self.tasks:
            wdir = os.path.dirname(l_path)
            bar.set_description(f"{wdir}: {os.path.basename(r_path)} and {os.path.basename(l_path)}")
            self.calc_interaction_from_dlg(r_path, l_path, method, mode, self.args.cutoff,
                                           getattr(self, f'output_fromater_{self.args.output_style}'),
                                           self.args.hydrogen_atom_only)
            bar.update(1)

_str2func = {
    'simple-analysis': simple_analysis,
}


def main(sys_args: List[str] = None):
    args_paser = argparse.ArgumentParser()
    subparsers = args_paser.add_subparsers(title='subcommands', dest='sub_command')
    simple_analysis_args = simple_analysis.make_args(subparsers.add_parser('simple-analysis', description='perform simple analysis on docking result'))

    args = args_paser.parse_args(sys_args)
    if args.sub_command in _str2func:
        _str2func[args.sub_command](args).excute()


if __name__ == "__main__":
    # main(r'simple-analysis -r receptor.pdbqt -l dock.pdbqt -bd data_tmp/docking --method pymol --mode '.split() + ['polar contact'])
    
    main()


'''
Date: 2024-12-13 23:07:00
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-12-25 20:50:42
Description: 
'''
import argparse
from pathlib import Path
from typing import List

from mbapy_lite.file import get_paths_with_extension
from pymol import cmd
from tqdm import tqdm

if __name__ == '__main__':
    from lazydock.pml.align_to_axis import align_pose_to_axis
    from lazydock.scripts._script_utils_ import clean_path, show_args
else:
    from ..pml.align_to_axis import align_pose_to_axis
    from ._script_utils_ import clean_path, show_args
    

def main(sys_args: List[str] = None):
    args_paser = argparse.ArgumentParser(description = 'Get pocket residues from ProteinPlus web server.')
    args_paser.add_argument('-d', '--dir', type = str, default='.',
                            help='input dir. Default is %(default)s.')
    args_paser.add_argument('-n', '--name', type = str, default='',
                            help='sub string in file name. Default is %(default)s.')
    args_paser.add_argument('-t', '--type', type = str, default='pdb',
                            help='file type. Default is %(default)s.')
    args_paser.add_argument('-s', '--suffix', type=str, default='aligned_to_axis',
                            help='output suffix. Default is %(default)s.')
    args_paser.add_argument('-r', '--recursive', action='store_true',
                            help='search subdirectories recursively. Default is %(default)s.')
    args = args_paser.parse_args(sys_args)
    # process IO path
    args.dir = clean_path(args.dir)
    args.type = args.type.split(',') if ',' in args.type else [args.type]
    paths = get_paths_with_extension(args.dir, args.type, recursive=args.recursive, name_substr=args.name)
    if not paths:
        raise ValueError(f'No file found in {args.receptor}.')
    # show args
    show_args(args, ['dir', 'type', 'suffix', 'recursive'])
    # sort bonds
    for path in tqdm(paths, total=len(paths)):
        path = Path(path)
        cmd.reinitialize()
        cmd.load(str(path), 'mol')
        align_pose_to_axis('mol')
        cmd.save(f'{path.parent}/{path.stem}_{args.suffix}.{path.suffix}', 'mol')


if __name__ == "__main__":
    main()


'''
Date: 2024-12-04 20:58:39
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-12-17 22:56:54
Description: 
'''

import argparse
import os
import time
from functools import wraps
from pathlib import Path
from typing import Dict, List, Tuple, Union

from mbapy_lite.base import put_err, Configs
from mbapy_lite.file import get_paths_with_extension, opts_file
from mbapy_lite.web import TaskPool, random_sleep
from tqdm import tqdm

from lazydock.pml.autodock_utils import DlgFile
from lazydock.scripts._script_utils_ import Command, clean_path, excute_command
from lazydock.web.hdock import run_dock_on_HDOCK, run_dock_on_HPEPDOCK


class vina(Command):
    def __init__(self, args, printf = print):
        super().__init__(args, printf)
        self.taskpool = None
        
    @staticmethod
    def make_args(args: argparse.ArgumentParser):
        args.add_argument('-d', '--dir', type = str, default='.',
                                help='config file directory contains config files (named "config.txt"), each sub-directory is a task. Default is %(default)s.')
        args.add_argument('-v', '--vina-name', type = str, default='vina',
                                help='vina executable name to call. Default is %(default)s.')
        args.add_argument('-n', '--n-workers', type=int, default=1,
                                help='number of tasks to parallel docking. Default is %(default)s.')
        return args
    
    def process_args(self):
        self.args.dir = clean_path(self.args.dir)
        if self.args.n_workers <= 0:
            put_err(f'n_workers must be positive integer, got {self.args.n_workers}, exit.', _exit=True)
        self.taskpool = TaskPool('threads', self.args.n_workers).start()
        
    @staticmethod
    def run_vina(config_path: Path, vina_name: str):
        print(f'current: {config_path}')
        if (config_path.parent / 'dock.pdbqt').exists():
            print(f'{config_path.parent} has done, skip')
            return 
        os.system(f'cd "{config_path.absolute().parent}" && {vina_name} --config ./config.txt --log ./log.txt')
        
    def main_process(self):
        if os.path.isdir(self.args.dir):
            configs_path = get_paths_with_extension(self.args.dir, ['.txt'], name_substr='config')
        else:
            put_err(f'dir argument should be a directory: {self.args.config}, exit.', _exit=True)
        print(f'get {len(configs_path)} config(s) for docking')
        tasks = []
        for config_path in tqdm(configs_path, total=len(configs_path)):
            tasks.append(self.taskpool.add_task(None, self.run_vina, Path(config_path), self.args.vina_name))
            while self.taskpool.count_waiting_tasks() > 1:
                time.sleep(1)
        self.taskpool.wait_till_tasks_done(tasks)
        self.taskpool.close()
        
        
def hdock_run_fn_warpper(result_prefix: str = 'HDOCK'):
    def ret_warpper(func):
        def core_wrapper(*args, **kwargs):
            config_path = args[0] if len(args) > 0 else kwargs.get('config_path', None)
            if config_path is None:
                put_err('config_path is required, exit.', _exit=True)
            # get parameters from config file
            if isinstance(config_path, Path):
                root = config_path.parent
                parameters = hdock.get_paramthers_from_config(config_path)
                parameters['receptor_path'] = config_path.parent / parameters['receptor']
                parameters['ligand_path'] = config_path.parent / parameters['ligand']
            # OR get parameters from receptor and ligand path in tuple
            elif isinstance(config_path, tuple):
                root = Path(config_path[0]).parent
                parameters = {'receptor_path': config_path[0], 'ligand_path': config_path[1]}
            # un-expected type
            else:
                put_err(f'config_path type not support: {type(config_path)}, exit.', _exit=True)
            # check if done
            if (root / f'{result_prefix}_all_results.tar.gz').exists():
                return print(f'{root} has done, skip')
            # perform docking
            print(f'current: {root}')
            ret =  func(*args, parameters=parameters, **kwargs)
            return ret
        return core_wrapper
    return ret_warpper


class hdock(Command):
    def __init__(self, args, printf = print):
        super().__init__(args, printf)
    
    @staticmethod
    def make_args(args: argparse.ArgumentParser):
        args.add_argument('-d', '--dir', type = str, default='.',
                          help='vina config file directory contains config files (named "config.txt"), each sub-directory is a task. Default is %(default)s.')
        args.add_argument('-r', '--receptor', type = str, default=None,
                          help="receptor pdb file name, optional. If provided, will ignore config.txt.")
        args.add_argument('-l', '--ligand', type = str, default=None,
                          help="ligand pdb file name, optional. If provided, will ignore config.txt.")
        args.add_argument('-m', '--method', type = str, default='web', choices=['web'],
                          help='docking method. Currently support "web". Default is %(default)s.')
        args.add_argument('--email', type = str, default=None,
                          help='email address for HDOCK web server. Default is %(default)s.')
        args.add_argument('-gui', '--gui', action='store_true', default=False,
                          help='show browser GUI. Default is %(default)s.')
        return args
    
    @staticmethod
    def get_paramthers_from_config(config_path: Path) -> Dict:
        config_lines = config_path.read_text().split('\n')
        paramthers = {line.split('=')[0].strip():line.split('=')[1].strip() for line in config_lines if not line.startswith('#')}
        return {k:v[:v.find('#')] for k,v in paramthers.items()}

    @staticmethod
    @hdock_run_fn_warpper('HDOCK')
    def run_hdock_web(config_path: Union[Path, Tuple[str, str]], parameters: Dict[str, str] = None, email=None, browser=None):
        w_dir = config_path.parent if isinstance(config_path, Path) else Path(config_path[0]).parent
        run_dock_on_HDOCK(receptor_path=parameters['receptor_path'], ligand_path=parameters['ligand_path'],
                          w_dir=w_dir, email=email, browser=browser)

    @staticmethod
    @hdock_run_fn_warpper('HDOCK')
    def run_hdock_local(config_path: Union[Path, Tuple[str, str]], parameters: Dict[str, str] = None, **kwargs):
        raise NotImplementedError('local docking not implemented yet.')

    def main_process(self):
        if not os.path.isdir(self.args.dir):
            put_err(f'dir argument should be a directory: {self.args.config}, exit.', _exit=True)
        if self.args.receptor is not None and self.args.ligand is not None:
            r_paths = get_paths_with_extension(self.args.dir, [], name_substr=self.args.receptor)
            l_paths = get_paths_with_extension(self.args.dir, [], name_substr=self.args.ligand)
            if len(r_paths) != len(l_paths):
                r_roots = [os.path.dirname(p) for p in r_paths]
                l_roots = [os.path.dirname(p) for p in l_paths]
                roots_count = {root: r_roots.count(root)+l_roots.count(root) for root in (set(r_roots) | set(l_roots))}
                invalid_roots = '\n'.join([root for root, count in roots_count.items() if count != 2])
                return put_err(f"The number of receptor and ligand files is not equal, please check the input files.\ninvalid roots:\n{invalid_roots}")
            configs_path = [(r, l) for r, l in zip(r_paths, l_paths)]
        else:
            configs_path = get_paths_with_extension(self.args.dir, ['.txt'], name_substr='config')
        print(f'get {len(configs_path)} config(s) for docking')
        # allow browser gui
        if self.args.gui:
            from mbapy_lite.web import Browser
            browser = Browser(options=[f"--user-agent={Configs.web.chrome_driver_path}"])
        else:
            browser = None
        # perform docking
        dock_fn = getattr(self, f'run_hdock_{self.args.method}')
        for config_path in tqdm(configs_path, total=len(configs_path)):
            dock_fn(Path(config_path) if isinstance(config_path, str) else config_path, email=self.args.email, browser=browser)
            random_sleep(300, 180) # sleep 3~5 minutes to avoid overloading the server


class hpepdock(hdock):
    def __init__(self, args, printf = print):
        super().__init__(args, printf)
        
    @staticmethod
    def make_args(args: argparse.ArgumentParser):
        args = hdock.make_args(args)
        # make sure hpepdock only support web method
        method_arg_idx = args._actions.index(list(filter(lambda x: x.dest =='method', args._actions))[0])
        args._actions[method_arg_idx].choices = ['web']
        args._actions[method_arg_idx].default = 'web'
        args._actions[method_arg_idx].help = 'docking method. Currently support "web". Default is %(default)s.'
        return args
    
    @staticmethod
    @hdock_run_fn_warpper('HPEPDOCK')
    def run_hdock_web(config_path: Path, parameters: Dict[str, str] = None, email=None, browser=None):
        w_dir = config_path.parent if isinstance(config_path, Path) else Path(config_path[0]).resolve().parent
        run_dock_on_HPEPDOCK(receptor_path=parameters['receptor_path'], ligand_path=parameters['ligand_path'],
                             w_dir=w_dir, email=email, browser=browser)


def convert_result_run_convert(input_path: Path, output_path: Path, method: str):
    if method in {'lazydock', 'obabel'}:
        getattr(convert_result, f'run_convert_{method}')(input_path, output_path)
    else:
        put_err(f'unsupported convert method: {method}, exit.', _exit=True)


class convert_result(Command):
    def __init__(self, args, printf = print):
        super().__init__(args, printf)
        self.taskpool = None

    @staticmethod
    def make_args(args: argparse.ArgumentParser):
        args.add_argument('-d', '--dir', type = str, default='.',
                                help='input directory. Default is %(default)s.')
        args.add_argument('-n', '--name', type = str, default='',
                          help='input file name. Default is %(default)s.')
        args.add_argument('-i', '--input-type', type = str, default='pdbqt,dlg',
                          help='input file type. Default is %(default)s.')
        args.add_argument('-o', '--output-type', type = str, default='pdb', choices=['pdb'],
                          help='input file type. Default is %(default)s.')
        args.add_argument('-s', '--suffix', type = str, default='',
                          help='output file suffix. Default is %(default)s.')
        args.add_argument('-m', '--method', type = str, default='lazydock', choices=['lazydock', 'obabel'],
                          help='convert tools to use. Currently support "lazydock, obabel". Default is %(default)s.')
        args.add_argument('--n-workers', type=int, default=4,
                          help='number of tasks to parallel docking. Default is %(default)s.')
        return args
    
    def process_args(self):
        self.args.dir = clean_path(self.args.dir)
        self.args.input_type = self.args.input_type.split(',')
        if self.args.n_workers <= 0:
            put_err(f'n_workers must be positive integer, got {self.args.n_workers}, exit.', _exit=True)
        self.taskpool = TaskPool('process', self.args.n_workers).start()
        
    @staticmethod
    def run_convert_lazydock(input_path: Path, output_path: Path):
        dlg = DlgFile(input_path)
        pdb_string = '\n'.join([f'MODEL {i+1}\n'+pose.as_pdb_string().replace('\r\n', '\n')+'ENDMDL' for i, pose in enumerate(dlg.pose_lst)])
        opts_file(output_path, 'w', data=pdb_string)

    @staticmethod
    def run_convert_obabel(input_path: Path, output_path: Path):
        ty1, ty2 = input_path.suffix.lower()[1:], output_path.suffix.lower()[1:]
        os.system(f'obabel -i{ty1} "{str(input_path)}" -o{ty2} -O "{str(output_path)}"')

    def main_process(self):
        input_paths = get_paths_with_extension(self.args.dir, self.args.input_type, name_substr=self.args.name)
        print(f'get {len(input_paths)} input(s) for convert:\n', '\n'.join([f'{i+1}. {x}' for i, x in enumerate(input_paths)]))
        if input('start convert? (y/n) ').lower() != 'y':
            return 
        tasks = []
        for input_path in tqdm(input_paths, total=len(input_paths)):
            input_path = Path(input_path)
            output_path = input_path.parent / f'{input_path.stem}{self.args.suffix}.{self.args.output_type}'
            tasks.append(self.taskpool.add_task(None, convert_result_run_convert, input_path, output_path, self.args.method))
            while self.taskpool.count_waiting_tasks() > 0:
                time.sleep(1)
        self.taskpool.wait_till_tasks_done(tasks)
        self.taskpool.close()


_str2func = {
    'vina': vina,
    'hdock': hdock,
    'hpepdock': hpepdock,
    'convert-result': convert_result,
}

def main(sys_args: List[str] = None):
    args_paser = argparse.ArgumentParser(description = 'perform docking with Vina or other docking software.')
    subparsers = args_paser.add_subparsers(title='subcommands', dest='sub_command')

    vina_args = vina.make_args(subparsers.add_parser('vina', description='perform vina molecular docking.'))
    hdock_args = hdock.make_args(subparsers.add_parser('hdock', description='perform HDOCK molecular docking.'))
    hpepdock_args = hpepdock.make_args(subparsers.add_parser('hpepdock', description='perform HPEPDOCK molecular docking.'))
    convert_result_args = convert_result.make_args(subparsers.add_parser('convert-result', description='convert docking result file to pdb format.'))

    excute_command(args_paser, sys_args, _str2func)


if __name__ == "__main__":
    # main('convert-result -d data_tmp/docking/ligand1 -m lazydock --n-workers 1'.split())
    # main('hpepdock -d data_tmp/docking/ligand2 -m web -r rec.pdb -l ligand.pdb'.split())
    
    main()


'''
Date: 2024-11-13 22:01:55
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-11-14 16:33:22
Description: 
'''
import os
from typing import Dict, List, Union

import requests
from mbapy_lite.base import put_err
from mbapy_lite.game import BaseInfo


class Residue(BaseInfo):
    def __init__(self, sequence_number: int = 0, amino_acid: str = '', protein_segment: str = '',
                 display_generic_number: str = '', alternative_generic_numbers: List[Dict[str, Union[int, str]]] = None, **kwargs):
        super().__init__()
        self.idx = sequence_number
        self.name = amino_acid
        self.seg = protein_segment
        self.generic_number = display_generic_number
        self.alternative_generic_numbers = alternative_generic_numbers or []


class Protein(BaseInfo):
    def __init__(self, entry_name: str = '', from_cache: bool = True):
        super().__init__()
        # try to load from cache first if from_cache is True and cache exists
        self.cache_path = os.path.expanduser(f'~/.lazydock/cache/gpcrdb/{entry_name}.json')
        loaded_from_cache = False
        if from_cache and os.path.exists(self.cache_path):
            try:
                self.from_json(self.cache_path)
                loaded_from_cache = True
            except Exception as e:
                put_err(f"Error loading cache: {e}")
        # if not loaded_from_cache, load from web
        if not loaded_from_cache:
            req = requests.get(f"https://gpcrdb.org/services/residues/extended/{entry_name}/")
            self.req = req.json()
            self.residuies = list(map(lambda x: Residue(**x), self.req))
        # dump to cache if not loaded from cache
        if not loaded_from_cache:
            self.to_json(self.cache_path) # makedirs inner this func


if __name__ == '__main__':
    p = Protein('cnr1_human')
    print(p.residuies)


'''
Date: 2024-12-15 19:25:42
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-12-22 20:46:09
Description: 
'''
import os
from typing import List, Union

from pymol import cmd


def start_lazydock_server(host: str = 'localhost', port: int = 8085, quiet: int = 1):
    from lazydock.pml.server import VServer
    print(f'Starting LazyDock server on {host}:{port}, quiet={quiet}')
    VServer(host, port, not bool(quiet))

cmd.extend('start_lazydock_server', start_lazydock_server)


def align_pose_to_axis_warp(pml_name: str, move_name: str = None, fixed: Union[List[float], str] = 'center', state: int = 0, move_method: str = 'rotate', dss: int = 1, quiet: int = 0):
    from lazydock.pml.align_to_axis import align_pose_to_axis
    print('try drag coords or matrix manually before running this command will get secondary structure remained.')
    align_pose_to_axis(pml_name, move_name, fixed, state, move_method, dss, quiet)

cmd.extend('align_pose_to_axis', align_pose_to_axis_warp)


def open_vina_config_as_box(config_path: str, spacing: float = 1.0):
    from mbapy_lite.file import opts_file

    from lazydock.pml.thirdparty.draw_bounding_box import draw_box
    if not os.path.exists(config_path):
        return print(f'Config file {config_path} not found, skip.')
    cfg = opts_file(config_path, way='lines')
    get_line = lambda n: [line for line in cfg if line.startswith(n)][0]
    center = {line.split('=')[0].strip(): float(line.split('=')[1].strip()) for line in map(get_line, ['center_x', 'center_y', 'center_z'])}
    size = {line.split('=')[0].strip(): float(line.split('=')[1].strip()) for line in map(get_line, ['size_x', 'size_y', 'size_z'])}
    print(f'center: {center}, size: {size}')
    minz, miny, minz = [(center[f'center_{k}'] - size[f'size_{k}'] / 2) * spacing for k in ['x', 'y', 'z']]
    maxz, maxy, maxz = [(center[f'center_{k}'] + size[f'size_{k}'] / 2) * spacing for k in ['x', 'y', 'z']]
    draw_box(minz, miny, minz, maxz, maxy, maxz)
    
cmd.extend('open_vina_config_as_box', open_vina_config_as_box)


print('LazyDock plugin loaded.')
print('''
Commands (python API):
    start_lazydock_server(host='localhost', port=8085, quiet=1)
    align_pose_to_axis(pml_name, move_name='', fixed='center', state=0, move_method='rotate', dss=1, quite=0)
    open_vina_config_as_box(config_path, spacing=1.0)
''')